"""
This module defines classes and functions to work with SpaceWire packets.
"""

import logging
import os
import struct
from enum import IntEnum
from typing import Tuple
from typing import Union

import numpy as np

from egse.bits import clear_bit
from egse.bits import crc_calc
from egse.bits import set_bit
from egse.exceptions import Error
from egse.setup import SetupError
from egse.state import GlobalState

MODULE_LOGGER = logging.getLogger(__name__)

try:
    _ = os.environ["PLATO_CAMERA_IS_EM"]
    TWOS_COMPLEMENT_OFFSET = 32768 if _.capitalize() in ("1", "True", "Yes") else 0
except KeyError:
    TWOS_COMPLEMENT_OFFSET = 0

# RMAP Error Codes and Constants -------------------------------------------------------------------

RMAP_PROTOCOL_ID = 0x01
RMAP_TARGET_LOGICAL_ADDRESS_DEFAULT = 0xFE
RMAP_TARGET_KEY = 0xD1

# Error and Status Codes

RMAP_SUCCESS = 0
RMAP_GENERAL_ERROR = 1
RMAP_UNUSED_PACKET_TYPE_COMMAND_CODE = 2
RMAP_INVALID_KEY = 3
RMAP_INVALID_DATA_CRC = 4
RMAP_EARLY_EOP = 5
RMAP_TOO_MUCH_DATA = 6
RMAP_EEP = 7
RMAP_RESERVED = 8
RMAP_VERIFY_BUFFER_OVERRUN = 9
RMAP_NOT_IMPLEMENTED_AUTHORISED = 10
RMAP_RMW_DATA_LENGTH_ERROR = 11
RMAP_INVALID_TARGET_LOGICAL_ADDRESS = 12

# Memory Map layout --------------------------------------------------------------------------------

# NOTE: These memory areas are currently equal for N-FEE and F-FEE. Don't know if this will
#       change in the future.

CRITICAL_AREA_START = 0x0000_0000
CRITICAL_AREA_END = 0x0000_00FC
GENERAL_AREA_START = 0x0000_0100
GENERAL_AREA_END = 0x0000_06FC
HK_AREA_START = 0x0000_0700
HK_AREA_END = 0x0000_07FC
WINDOWING_AREA_START = 0x0080_0000
WINDOWING_AREA_END = 0x00FF_FFFC


class RMAPError(Error):
    """An RMAP specific Error."""

    pass


class CheckError(RMAPError):
    """
    Raised when a check fails and you want to pass a status values along with the message.
    """

    def __init__(self, message, status):
        self.message = message
        self.status = status


def update_transaction_identifier(tid: int) -> int:
    """
    Updates the transaction identifier and returns the new value.

    FIXME: document more about this identifier, where is it used, when is it checked,
           when does it need to be incremented, who initializes the identifier and
           who updates it, ...

    Args:
        tid (int): The current transaction identifier

    Returns:
        the updated transaction identifier (int).
    """
    tid = (tid + 1) & 0xFFFF
    return tid


def create_rmap_read_request_packet(address: int, length: int, tid: int, strict: bool = True) -> bytes:
    """
    Creates an RMAP Read Request SpaceWire packet.

    The read request is an RMAP command that read a number of bytes from the FEE register memory.

    The function returns a ``ctypes`` character array (which is basically a bytes array) that
    can be passed into the EtherSpaceLink library function ``esl_write_packet()``.

    Address shall be within the 0x0000_0000 and 0x00FF_FFFC. The memory map (register) is divided
    in the following areas:

        0x0000_0000 - 0x0000_00FC   Critical Configuration Area (verified write)
        0x0000_0100 - 0x0000_06FC   General Configuration Area (unverified write)
        0x0000_0700 - 0x0000_07FC   Housekeeping area
        0x0000_0800 - 0x007F_FFFC   Not Supported
        0x0080_0000 - 0x00FF_FFFC   Windowing Area (unverified write)
        0x0010_0000 - 0xFFFF_FFFC   Not Supported

    All read requests to the critical area shall have a fixed data length of 4 bytes.
    All read requests to a general area shall have a maximum data length of 256 bytes.
    All read requests to the housekeeping area shall have a maximum data length of 256 bytes.
    All read requests to the windowing area shall have a maximum data length of 4096 bytes.

    The transaction identifier shall be incremented for each read request. This shall be done by
    the calling function!

    Args:
        address (int): the FEE register memory address
        length (int): the data length
        tid (int): transaction identifier
        strict (bool): perform strict checking of address and length

    Returns:
        a bytes object containing the full RMAP Read Request packet.
    """

    check_address_and_data_length(address, length, strict=strict)

    buf = bytearray(16)

    # NOTE: The first bytes would each carry the target SpW address or a destination port,
    #       but this is not used for point-to-point connections, so we're safe.

    buf[0] = 0x51  # Target N-FEE or F-FEE
    buf[1] = 0x01  # RMAP Protocol ID
    buf[2] = 0x4C  # Instruction: 0b1001100, RMAP Request, Read, Incrementing address, reply address
    buf[3] = 0xD1  # Destination Key
    buf[4] = 0x50  # Initiator is always the DPU
    buf[5] = (tid >> 8) & 0xFF  # MSB of the Transition ID
    buf[6] = tid & 0xFF  # LSB of the Transition ID
    buf[7] = 0x00  # Extended address is not used
    buf[8] = (address >> 24) & 0xFF  # address (MSB)
    buf[9] = (address >> 16) & 0xFF  # address
    buf[10] = (address >> 8) & 0xFF  # address
    buf[11] = address & 0xFF  # address (LSB)
    buf[12] = (length >> 16) & 0xFF  # data length (MSB)
    buf[13] = (length >> 8) & 0xFF  # data length
    buf[14] = length & 0xFF  # data length (LSB)
    buf[15] = rmap_crc_check(buf, 0, 15) & 0xFF
    return bytes(buf)


def create_rmap_read_request_reply_packet(
    instruction_field: int, tid: int, status: int, buffer: bytes, buffer_length: int
) -> bytes:
    """
    Creates an RMAP Reply to a RMAP Read Request packet.

    The function returns a ``ctypes`` character array (which is basically a bytes array) that
    can be passed into the EtherSpaceLink library function ``esl_write_packet()``.

    Args:
        instruction_field (int): the instruction field of the RMAP read request packet
        tid (int): the transaction identifier of the read request packet
        status (int): the status field, 0 on success
        buffer (bytes): the data that was read as indicated by the read request
        buffer_length (int): the data length

    Returns:
        packet: a bytes object containing the full RMAP Reply packet.
    """

    buf = bytearray(12 + buffer_length + 1)

    buf[0] = 0x50  # Initiator address N-DPU or F-DPU
    buf[1] = 0x01  # RMAP Protocol ID
    buf[2] = instruction_field & 0x3F  # Clear the command bit as this is a reply
    buf[3] = status & 0xFF  # Status field: 0 on success
    buf[4] = 0x51  # Target address is always the N-FEE or F-FEE
    buf[5] = (tid >> 8) & 0xFF  # MSB of the Transition ID
    buf[6] = tid & 0xFF  # LSB of the Transition ID
    buf[7] = 0x00  # Reserved
    buf[8] = (buffer_length >> 16) & 0xFF  # data length (MSB)
    buf[9] = (buffer_length >> 8) & 0xFF  # data length
    buf[10] = buffer_length & 0xFF  # data length (LSB)
    buf[11] = rmap_crc_check(buf, 0, 11) & 0xFF  # Header CRC

    # Note that we assume here that len(buffer) == buffer_length.

    if len(buffer) != buffer_length:
        MODULE_LOGGER.warning(
            f"While creating an RMAP read reply packet, the length of the buffer ({len(buffer)}) "
            f"not equals the buffer_length ({buffer_length})"
        )

    for idx, value in enumerate(buffer):
        buf[12 + idx] = value

    buf[12 + buffer_length] = rmap_crc_check(buffer, 0, buffer_length) & 0xFF  # data CRC

    return bytes(buf)


def create_rmap_verified_write_packet(address: int, data: bytes, tid: int) -> bytes:
    """
    Create an RMAP packet for a verified write request on the FEE. The length of the data is
    by convention always 4 bytes and therefore not passed as an argument.

    Args:
        address: the start memory address on the FEE register map
        data: the data to be written in the register map at address [4 bytes]
        tid (int): transaction identifier

    Returns:
        packet: a bytes object containing the SpaceWire packet.
    """

    if len(data) < 4:
        raise ValueError(f"The data argument should be at least 4 bytes, but it is only {len(data)} bytes: {data=}.")

    if address > CRITICAL_AREA_END:
        raise ValueError("The address range for critical configuration is [0x00 - 0xFC].")

    tid = update_transaction_identifier(tid)

    # Buffer length is fixed at 24 bytes since the data length is fixed
    # at 4 bytes (32 bit addressing)

    buf = bytearray(21)

    # The values below are taken from the PLATO N-FEE to N-DPU
    # Interface Requirements Document [PLATO-DLR-PL-ICD-0010]

    buf[0] = 0x51  # Logical Address
    buf[1] = 0x01  # Protocol ID
    buf[2] = 0x7C  # Instruction
    buf[3] = 0xD1  # Key
    buf[4] = 0x50  # Initiator Address
    buf[5] = (tid >> 8) & 0xFF  # MSB of the Transition ID
    buf[6] = tid & 0xFF  # LSB of the Transition ID
    buf[7] = 0x00  # Extended address
    buf[8] = (address >> 24) & 0xFF  # address (MSB)
    buf[9] = (address >> 16) & 0xFF  # address
    buf[10] = (address >> 8) & 0xFF  # address
    buf[11] = address & 0xFF  # address (LSB)
    buf[12] = 0x00  # data length (MSB)
    buf[13] = 0x00  # data length
    buf[14] = 0x04  # data length (LSB)
    buf[15] = rmap_crc_check(buf, 0, 15) & 0xFF  # header CRC
    buf[16] = data[0]
    buf[17] = data[1]
    buf[18] = data[2]
    buf[19] = data[3]
    buf[20] = rmap_crc_check(buf, 16, 4) & 0xFF  # data CRC

    return bytes(buf)


def create_rmap_unverified_write_packet(address: int, data: bytes, length: int, tid: int) -> bytes:
    """
    Create an RMAP packet for a unverified write request on the FEE.

    Args:
        address: the start memory address on the FEE register map
        data: the data to be written in the register map at address
        length: the length of the data
        tid (int): transaction identifier

    Returns:
        packet: a bytes object containing the SpaceWire packet.
    """

    # We can only handle data for which the length >= the given length argument.

    if len(data) < length:
        raise ValueError(
            f"The length of the data argument ({len(data)}) is smaller than the given length argument ({length})."
        )

    if len(data) > length:
        MODULE_LOGGER.warning(
            f"The length of the data argument ({len(data)}) is larger than "
            f"the given length argument ({length}). The data will be truncated "
            f"when copied into the packet."
        )

    if address <= CRITICAL_AREA_END:
        raise ValueError(
            f"The given address (0x{address:08X}) is in the range for critical configuration is "
            f"[0x00 - 0xFC]. Use the verified write function for this."
        )

    tid = update_transaction_identifier(tid)

    # Buffer length is fixed at 24 bytes since the data length
    # is fixed at 4 bytes (32 bit addressing)

    buf = bytearray(16 + length + 1)
    offset = 0

    buf[offset + 0] = 0x51  # Logical Address
    buf[offset + 1] = 0x01  # Protocol ID
    buf[offset + 2] = 0x6C  # Instruction
    buf[offset + 3] = 0xD1  # Key
    buf[offset + 4] = 0x50  # Initiator Address
    buf[offset + 5] = (tid >> 8) & 0xFF  # MSB of the Transition ID
    buf[offset + 6] = tid & 0xFF  # LSB of the Transition ID
    buf[offset + 7] = 0x00  # Extended address
    buf[offset + 8] = (address >> 24) & 0xFF  # address (MSB)
    buf[offset + 9] = (address >> 16) & 0xFF  # address
    buf[offset + 10] = (address >> 8) & 0xFF  # address
    buf[offset + 11] = address & 0xFF  # address (LSB)
    buf[offset + 12] = (length >> 16) & 0xFF  # data length (MSB)
    buf[offset + 13] = (length >> 8) & 0xFF  # data length
    buf[offset + 14] = length & 0xFF  # data length (LSB)
    buf[offset + 15] = rmap_crc_check(buf, 0, 15) & 0xFF  # header CRC

    offset += 16

    for idx, value in enumerate(data):
        buf[offset + idx] = value

    buf[offset + length] = rmap_crc_check(buf, offset, length) & 0xFF  # data CRC

    return bytes(buf)


def create_rmap_write_request_reply_packet(instruction_field: int, tid: int, status: int) -> bytes:
    buf = bytearray(8)

    buf[0] = 0x50  # Initiator address N-DPU or F-DPU
    buf[1] = 0x01  # RMAP Protocol ID
    buf[2] = instruction_field & 0x3F  # Clear the command bit as this is a reply
    buf[3] = status & 0xFF  # Status field: 0 on success
    buf[4] = 0x51  # Target address is always the N-FEE or F-FEE
    buf[5] = (tid >> 8) & 0xFF  # MSB of the Transition ID
    buf[6] = tid & 0xFF  # LSB of the Transition ID
    buf[7] = rmap_crc_check(buf, 0, 7) & 0xFF  # Header CRC

    return bytes(buf)


def check_address_and_data_length(address: int, length: int, strict: bool = True) -> None:
    """
    Checks the address and length in the range of memory areas used by the FEE.

    The ranges are taken from the PLATO-DLR-PL-ICD-0010 N-FEE to N-DPU IRD.

    Args:
        address (int): the memory address of the FEE Register
        length (int): the number of bytes requested
        strict (bool): strictly apply the rules

    Raises:
        RMAPError: when address + length fall outside any specified area.
    """

    if not strict:
        # All these restrictions have been relaxed on the N-FEE.
        # We are returning here immediately instead of removing or commenting out the code.
        # These reason is that we can then bring back restriction easier and gradually.

        MODULE_LOGGER.warning(
            "Address and data length checks have been disabled, because the N-FEE "
            "does not enforce restrictions in the critical memory area."
        )
        return

    if length % 4:
        raise RMAPError("The requested data length shall be a multiple of 4 bytes.", address, length)

    if address % 4:
        raise RMAPError("The address shall be a multiple of 4 bytes.", address, length)

    # Note that when checking the given data length, at the defined area end,
    # we can still read 4 bytes.

    if CRITICAL_AREA_START <= address <= CRITICAL_AREA_END:
        if length != 4:
            raise RMAPError("Read requests to the critical area have a fixed data length of 4 bytes.", address, length)
    elif GENERAL_AREA_START <= address <= GENERAL_AREA_END:
        if length > 256:
            raise RMAPError(
                "Read requests to the general area have a maximum data length of 256 bytes.", address, length
            )
        if address + length > GENERAL_AREA_END + 4:
            raise RMAPError(
                "The requested data length for the general area is too large. "
                "The address + length exceeds the general area boundaries.",
                address,
                length,
            )

    elif HK_AREA_START <= address <= HK_AREA_END:
        if length > 256:
            raise RMAPError(
                "Read requests to the housekeeping area have a maximum data length of 256 bytes.", address, length
            )
        if address + length > HK_AREA_END + 4:
            raise RMAPError(
                "The requested data length for the housekeeping area is too large. "
                "The address + length exceeds the housekeeping area boundaries.",
                address,
                length,
            )

    elif WINDOWING_AREA_START <= address <= WINDOWING_AREA_END:
        if length > 4096:
            raise RMAPError(
                "Read requests to the windowing area have a maximum data length of 4096 bytes.", address, length
            )
        if address + length > WINDOWING_AREA_END + 4:
            raise RMAPError(
                "The requested data length for the windowing area is too large. "
                "The address + length exceeds the windowing area boundaries.",
                address,
                length,
            )

    else:
        raise RMAPError("Register address for RMAP read requests is invalid.", address, length)


class PacketType(IntEnum):
    """Enumeration type that defines the SpaceWire packet type."""

    DATA_PACKET = 0
    OVERSCAN_DATA = 1
    HOUSEKEEPING_DATA = 2  # N-FEE
    DEB_HOUSEKEEPING_DATA = 2  # F-FEE
    AEB_HOUSEKEEPING_DATA = 3  # F-FEE


class DataPacketType:
    """
    Defines the Data Packet Field: Type, which is a bit-field of 16 bits.

    Properties:
      * value: returns the data type as an integer
      * packet_type: the type of data packet, defined in PacketType enum.
      * mode: the FEE mode, defined in n_fee_mode and f_fee_mode enum
      * last_packet: flag which defines the last packet of a type in the current readout cycle
      * ccd_side: 0 for E-side (left), 1 for F-side (right), see egse.fee.fee_side
      * ccd_number: CCD number [0, 3]
      * frame_number: the frame number after sync
    """

    def __init__(self, data_type: int = 0):
        self._data_type: int = data_type
        # self.n_fee_side = GlobalState.setup.camera.fee.ccd_sides.enum

    @property
    def value(self) -> int:
        """Returns the data packet type as an int."""
        return self._data_type

    @property
    def packet_type(self):
        """Returns the packet type: 0 = data packet, 1 = overscan data, 2 = housekeeping packet."""
        return self._data_type & 0b0011

    @packet_type.setter
    def packet_type(self, value):
        if not 0 <= value < 3:
            raise ValueError(f"Packet Type can only have the value 0, 1, or 2, {value=} given.")
        x = self._data_type
        for idx, bit in enumerate([0, 1]):
            x = set_bit(x, bit) if value & (1 << idx) else clear_bit(x, bit)
        self._data_type = x

    @property
    def mode(self) -> int:
        return (self._data_type & 0b1111_0000_0000) >> 8

    @mode.setter
    def mode(self, value: int):
        x = self._data_type
        for idx, bit in enumerate([8, 9, 10, 11]):
            x = set_bit(x, bit) if value & (1 << idx) else clear_bit(x, bit)
        self._data_type = x

    @property
    def last_packet(self) -> bool:
        return bool(self._data_type & 0b1000_0000)

    @last_packet.setter
    def last_packet(self, flag: bool):
        self._data_type = set_bit(self._data_type, 7) if flag else clear_bit(self._data_type, 7)

    @property
    def ccd_side(self) -> int:
        return (self._data_type & 0b0100_0000) >> 6

    @ccd_side.setter
    def ccd_side(self, value: int):
        self._data_type = set_bit(self._data_type, 6) if value & 0b0001 else clear_bit(self._data_type, 6)

    @property
    def ccd_number(self) -> int:
        return (self._data_type & 0b0011_0000) >> 4

    @ccd_number.setter
    def ccd_number(self, value):
        x = self._data_type
        for idx, bit in enumerate([4, 5]):
            x = set_bit(x, bit) if value & (1 << idx) else clear_bit(x, bit)
        self._data_type = x

    @property
    def frame_number(self) -> int:
        return (self._data_type & 0b1100) >> 2

    @frame_number.setter
    def frame_number(self, value):
        x = self._data_type
        for idx, bit in enumerate([2, 3]):
            x = set_bit(x, bit) if value & (1 << idx) else clear_bit(x, bit)
        self._data_type = x

    def __str__(self) -> str:
        try:
            from egse.fee import n_fee_mode

            mode = n_fee_mode(self.mode).name
        except ImportError:
            mode = str(self.mode)

        n_fee_side = GlobalState.setup.camera.fee.ccd_sides.enum

        return (
            f"mode:{mode}, last_packet:{self.last_packet}, "
            f"CCD side:{n_fee_side(self.ccd_side).name}, CCD number:{self.ccd_number}, "
            f"Frame number:{self.frame_number}, Packet Type:{PacketType(self.packet_type).name}"
        )


def to_string(data: Union[DataPacketType]) -> str:
    """Returns a 'user-oriented' string representation of the SpW DataPacketType.

    The purpose of this function is to represent the N-FEE information in a user-oriented way.
    That means for certain values that they will be converted into the form the a user understands
    and that may be different or reverse from the original N-FEE definition. An example is the
    CCD number which is different from the user perspective with respect to the N-FEE.

    If any other object type is passed, the data.__str__() method will be returned without
    processing or conversion.

    Args:
        data: a DataPacketType
    """
    try:
        from egse.fee import n_fee_mode

        mode = n_fee_mode(data.mode).name
    except ImportError:
        mode = str(data.mode)

    n_fee_side = GlobalState.setup.camera.fee.ccd_sides.enum

    if isinstance(data, DataPacketType):
        try:
            ccd_bin_to_id = GlobalState.setup.camera.fee.ccd_numbering.CCD_BIN_TO_ID
        except AttributeError:
            raise SetupError("No entry in the setup for camera.fee.ccd_numbering.CCD_BIN_TO_ID")
        return (
            f"mode:{mode}, last_packet:{data.last_packet}, "
            f"CCD side:{n_fee_side(data.ccd_side).name}, CCD number:"
            f"{ccd_bin_to_id[data.ccd_number]}, "
            f"Frame number:{data.frame_number}, Packet Type:{PacketType(data.packet_type).name}"
        )
    else:
        return data.__str__()


class DataPacketHeader:
    """
    Defines the header of a data packet.

    The full header can be retrieved as a bytes object with the `data_as_bytes()` method.

    Properties:
      * logical_address: fixed value of  0x50
      * protocol_id: fixed value of 0xF0
      * length: length of the data part of the packet, i.e. the packet length - size of the header
      * type: data packet type as defined by DataPacketType
      * frame_counter:
      * sequence_counter: a packet sequence counter per CCD
    """

    def __init__(self, header_data: bytes = None):
        self.header_data = bytearray(header_data or bytes([0x50, 0xF0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]))

        if len(self.header_data) != 10:
            raise ValueError(
                f"The length of the header for a data packet shall be 10 bytes, got {len(self.header_data)}."
            )

        self.n_fee_side = GlobalState.setup.camera.fee.ccd_sides.enum

    def data_as_bytes(self) -> bytes:
        """Returns the full header as a bytes object."""
        return bytes(self.header_data)

    @property
    def logical_address(self) -> int:
        return self.header_data[0]

    @logical_address.setter
    def logical_address(self, value: int):
        self.header_data[0] = value

    @property
    def protocol_id(self) -> int:
        return self.header_data[1]

    @protocol_id.setter
    def protocol_id(self, value: int):
        self.header_data[1] = value

    @property
    def length(self) -> int:
        return int.from_bytes(self.header_data[2:4], byteorder="big")

    @length.setter
    def length(self, value: int):
        self.header_data[2:4] = value.to_bytes(2, "big")

    @property
    def type(self):
        return int.from_bytes(self.header_data[4:6], byteorder="big")

    @type.setter
    def type(self, value: Union[int, bytes, DataPacketType]):
        if isinstance(value, bytes):
            self.header_data[4:6] = value
        elif isinstance(value, DataPacketType):
            self.header_data[4:6] = value.value.to_bytes(2, "big")
        else:
            self.header_data[4:6] = value.to_bytes(2, "big")

    @property
    def type_as_object(self):
        return DataPacketType(self.type)

    @property
    def packet_type(self):
        return self.type_as_object.packet_type

    @packet_type.setter
    def packet_type(self, value: int):
        type_obj = self.type_as_object
        type_obj.packet_type = value
        self.type = type_obj

    @property
    def last_packet(self):
        return self.type_as_object.last_packet

    @last_packet.setter
    def last_packet(self, flag: bool):
        type_obj = self.type_as_object
        type_obj.last_packet = flag
        self.type = type_obj

    @property
    def frame_counter(self):
        return int.from_bytes(self.header_data[6:8], byteorder="big")

    @frame_counter.setter
    def frame_counter(self, value):
        self.header_data[6:8] = value.to_bytes(2, "big")

    @property
    def sequence_counter(self):
        return int.from_bytes(self.header_data[8:10], byteorder="big")

    @sequence_counter.setter
    def sequence_counter(self, value):
        self.header_data[8:10] = value.to_bytes(2, "big")

    def as_dict(self):
        data_packet_type = DataPacketType(self.type)
        try:
            from egse.fee import n_fee_mode

            mode = n_fee_mode(data_packet_type.mode).name
        except ImportError:
            mode = str(data_packet_type.mode)

        return dict(
            logical_address=f"0x{self.logical_address:02X}",
            protocol_id=f"0x{self.protocol_id:02X}",
            length=self.length,
            type=f"0x{self.type:04X}",
            frame_counter=self.frame_counter,
            sequence_counter=self.sequence_counter,
            packet_type=data_packet_type.packet_type,
            frame_number=data_packet_type.frame_number,
            ccd_number=data_packet_type.ccd_number,
            ccd_side=self.n_fee_side(data_packet_type.ccd_side).name,
            last_packet=data_packet_type.last_packet,
            mode=mode,
        )


class SpaceWirePacket:
    """Base class for any packet transmitted over a SpaceWire cable."""

    # these settings are used by this class and its sub-classes to configure the print options
    # for the numpy arrays.

    _threshold = 300  # sys.maxsize
    _edgeitems = 10
    _linewidth = 120

    def __init__(self, data: Union[bytes, np.ndarray]):
        """
        Args:
            data: a bytes object or a numpy array of type np.uint8 (not enforced)
        """
        self._bytes = bytes(data)

    def __repr__(self):
        options = np.get_printoptions()
        np.set_printoptions(
            formatter={"int": lambda x: f"0x{x:02x}"},
            threshold=self._threshold,
            edgeitems=self._edgeitems,
            linewidth=self._linewidth,
        )
        msg = f"{self.__class__.__name__}({self._bytes})"
        np.set_printoptions(**options)
        return msg

    @property
    def packet_as_bytes(self):
        return self._bytes

    @property
    def packet_as_ndarray(self):
        return np.frombuffer(self._bytes, dtype=np.uint8)

    @property
    def logical_address(self):
        # TODO: what about a timecode, that has no logical address?
        return self._bytes[0]

    @property
    def protocol_id(self):
        # TODO: what about a timecode, that has no protocol id?
        return self._bytes[1]

    def header_as_bytes(self) -> bytes:
        # TODO: what about timecode, this has no header, except maybe the first byte: 0x91
        raise NotImplementedError

    @staticmethod
    def create_packet(data: Union[bytes, np.ndarray]):
        """
        Factory method that returns a SpaceWire packet of the correct type based on the information
        in the header.
        """
        if TimecodePacket.is_timecode_packet(data):
            return TimecodePacket(data)
        if HousekeepingPacket.is_housekeeping_packet(data):
            return HousekeepingPacket(data)
        if DataDataPacket.is_data_data_packet(data):
            return DataDataPacket(data)
        if OverscanDataPacket.is_overscan_data_packet(data):
            return OverscanDataPacket(data)
        if WriteRequest.is_write_request(data):
            return WriteRequest(data)
        if WriteRequestReply.is_write_reply(data):
            return WriteRequestReply(data)
        if ReadRequest.is_read_request(data):
            return ReadRequest(data)
        if ReadRequestReply.is_read_reply(data):
            return ReadRequestReply(data)
        return SpaceWirePacket(data)


class DataPacket(SpaceWirePacket):
    """
    Base class for proprietary SpaceWire data packets that are exchanged between FEE and DPU.

    .. note::
        This class should not be instantiated directly. Use the SpaceWirePacket.create_packet()
        factory method or the constructors of one of the sub-classes of this DataPacket class.
    """

    DATA_HEADER_LENGTH = 10

    def __init__(self, data: Union[bytes, np.ndarray]):
        """
        Args:
            data: a bytes object or a numpy array
        """
        if not self.is_data_packet(data):
            raise ValueError(f"Can not create a DataPacket from the given data {[f'0x{x:02x}' for x in data]}")

        super().__init__(data)

        if (data[2] == 0x00 and data[3] == 0x00) or len(data) == self.DATA_HEADER_LENGTH:
            MODULE_LOGGER.warning(f"SpaceWire data packet without data found, packet={[f'0x{x:02x}' for x in data]}")

        self._length = (data[2] << 8) + data[3]

        if len(data) != self._length + self.DATA_HEADER_LENGTH:
            MODULE_LOGGER.warning(
                f"The length of the data argument ({len(data)}) given to "
                f"the constructor of {self.__class__.__name__} (or sub-classes) is inconsistent "
                f"with the length data field ({self._length} + 10) in the packet header."
            )
            raise ValueError(
                f"{self.__class__.__name__} header: data-length field ({self._length}) not "
                f"consistent with packet length ({len(data)}). Difference should be "
                f"{self.DATA_HEADER_LENGTH}."
            )

        self._type = DataPacketType((data[4] << 8) + data[5])
        self._data = None  # lazy loading of data from self._bytes

    @property
    def length(self) -> int:
        """Returns the data length in bytes.

        .. note:: length == len(data_nd_array) * 2
            This length property returns the length of the data area in bytes. This value is
            taken  from the header of the data packet. If you want to compare this with the size
            of the data_as_ndarray property, multiply the length by 2 because the data is 16-bit
            integers, not bytes.

        Returns:
            the size of the data area of the packet in bytes.
        """
        return self._length

    @property
    def data_as_ndarray(self):
        """
        Returns the data from this data packet as a 16-bit integer Numpy array.

        .. note::
            The data has been converted from the 8-bit packet data into 16-bit integers. That
            means the length of this data array will be half the length of the data field the
            packet, i.e. ``len(data) == length // 2``.
            The reason for this is that pixel data has a size of 16-bit.

        .. todo::
            check if the data-length of HK packets should also be a multiple of 16.

        Returns:
            data: Numpy array with the data from this packet (type is np.uint16)

        """

        # We decided to lazy load/construct the data array. The reason is that the packet may be
        # created / transferred without the need to unpack the data field into a 16-bit numpy array.

        if self._data is None:
            # The data is in two's-complement. The most significant bit (msb) shall be inverted
            # according to Sampie Smit. That is done in the following line where the msb in each
            # byte on an even index is inverted.

            # data = [toggle_bit(b, 7) if not idx % 2 else b for idx, b in enumerate(self._bytes)]
            # data = bytearray(data)
            # data_1 = np.frombuffer(data, offset=10, dtype='>u2')

            # Needs further confirmation, but the following line should have the same effect as
            # the previous three lines.
            data_2 = np.frombuffer(self._bytes, offset=10, dtype=">i2") + TWOS_COMPLEMENT_OFFSET

            # Test if the results are identical, left the code in until we are fully confident
            # if diff := np.sum(np.cumsum(data_1 - data_2)):
            #     MODULE_LOGGER.info(f"cumsum={diff}")

            self._data = data_2.astype("uint16")
        return self._data

    @property
    def data(self) -> bytes:
        return self._bytes[10 : 10 + self._length]

    @property
    def type(self) -> DataPacketType:
        return self._type

    @property
    def frame_counter(self):
        return (self._bytes[6] << 8) + self._bytes[7]

    @property
    def sequence_counter(self):
        return (self._bytes[8] << 8) + self._bytes[9]

    @property
    def header(self) -> DataPacketHeader:
        return DataPacketHeader(self.header_as_bytes())

    def header_as_bytes(self):
        return self._bytes[:10]

    @classmethod
    def is_data_packet(cls, data: np.ndarray) -> bool:
        if len(data) < 10 or data[0] != 0x50 or data[1] != 0xF0:
            return False
        return True

    def __str__(self):
        options = np.get_printoptions()
        np.set_printoptions(
            formatter={"int": lambda x: f"0x{x:04x}"},
            threshold=super()._threshold,
            edgeitems=super()._edgeitems,
            linewidth=super()._linewidth,
        )
        msg = (
            f"{self.__class__.__name__}:\n"
            f"  Logical Address = 0x{self.logical_address:02X}\n"
            f"  Protocol ID = 0x{self.protocol_id:02X}\n"
            f"  Length = {self.length}\n"
            f"  Type = {self._type}\n"
            f"  Frame Counter = {self.frame_counter}\n"
            f"  Sequence Counter = {self.sequence_counter}\n"
            f"  Data = \n{self.data}"
        )
        np.set_printoptions(**options)
        return msg


class DataDataPacket(DataPacket):
    """Proprietary Data Packet for N-FEE and F-FEE CCD image data."""

    @classmethod
    def is_data_data_packet(cls, data: Union[bytes, np.ndarray]) -> bool:
        if len(data) <= 10:
            return False
        if data[0] != 0x50:
            return False
        if data[1] != 0xF0:
            return False
        type_ = DataPacketType((data[4] << 8) + data[5])
        if type_.packet_type == PacketType.DATA_PACKET:
            return True
        return False


class OverscanDataPacket(DataPacket):
    """Proprietary Overscan Data Packet for N-FEE and F-FEE CCD image data."""

    @classmethod
    def is_overscan_data_packet(cls, data: Union[bytes, np.ndarray]) -> bool:
        if len(data) <= 10:
            return False
        if data[0] != 0x50:
            return False
        if data[1] != 0xF0:
            return False
        type_ = DataPacketType((data[4] << 8) + data[5])
        if type_.packet_type == PacketType.OVERSCAN_DATA:
            return True
        return False


class HousekeepingPacket(DataPacket):
    """Proprietary Housekeeping data packet for the N-FEE and F-FEE."""

    def __init__(self, data: Union[bytes, np.ndarray]):
        """
        Args:
            data: a numpy array of type np.uint8 (not enforced)
        """
        if not self.is_housekeeping_packet(data):
            raise ValueError(f"Can not create a HousekeepingPacket from the given data {data}")

        # The __init__ method of DataPacket already checks e.g. data-length against packet length,
        # so there is no need for these tests here.

        super().__init__(data)

    @classmethod
    def is_housekeeping_packet(cls, data: Union[bytes, np.ndarray]) -> bool:
        if len(data) <= 10:
            return False
        if data[0] != 0x50:
            return False
        if data[1] != 0xF0:
            return False
        type_ = DataPacketType((data[4] << 8) + data[5])
        if type_.packet_type == PacketType.HOUSEKEEPING_DATA:
            return True
        return False


class TimecodePacket(SpaceWirePacket):
    """A Timecode Packet.

    This packet really is an extended packet which is generated by the Diagnostic SpaceWire
    Interface (DSI) to forward a SpaceWire timecode over the Ethernet connection.
    """

    def __init__(self, data: Union[bytes, np.ndarray]):
        super().__init__(data)

    @property
    def timecode(self) -> int:
        return self._bytes[1] & 0x3F

    def header_as_bytes(self) -> bytes:
        return self._bytes[0:1]

    @classmethod
    def is_timecode_packet(cls, data: Union[bytes, np.ndarray]) -> bool:
        return data[0] == 0x91

    def __str__(self):
        return f"Timecode Packet: timecode = 0x{self.timecode:x}"


class RMAPPacket(SpaceWirePacket):
    """Base class for RMAP SpaceWire packets."""

    def __init__(self, data: Union[bytes, np.ndarray]):
        if not self.is_rmap_packet(data):
            raise ValueError(f"Can not create a RMAPPacket from the given data {data}")
        super().__init__(data)

    def __str__(self):
        return f"{self.__class__.__name__}:\n  Logical Address = 0x{self.logical_address:02X}\n  Data = {self.data}\n"

    @property
    def instruction(self):
        return get_instruction_field(self._bytes)

    @property
    def transaction_id(self):
        return get_transaction_identifier(self._bytes)

    @classmethod
    def is_rmap_packet(cls, data: Union[bytes, np.ndarray]):
        if data[1] == 0x01:  # Protocol ID
            return True
        return False


class WriteRequest(RMAPPacket):
    """A Write Request SpaceWire RMAP Packet."""

    def __init__(self, data: Union[bytes, np.ndarray]):
        super().__init__(data)

    def is_verified(self):
        return self._bytes[2] == 0x7C

    def is_unverified(self):
        return self._bytes[2] == 0x6C

    @property
    def address(self):
        return get_address(self._bytes)

    @property
    def data_length(self):
        return get_data_length(self._bytes)

    @property
    def data(self) -> bytes:
        return get_data(self._bytes)

    @classmethod
    def is_write_request(cls, data: Union[bytes, np.ndarray]):
        if not RMAPPacket.is_rmap_packet(data):
            return False
        if data[0] != 0x51:
            return False
        if (data[2] == 0x7C or data[2] == 0x6C) and data[3] == 0xD1:
            return True
        return False

    def __str__(self):
        prefix = "Verified" if self.is_verified() else "Unverified"
        return f"{prefix} Write Request: {self.transaction_id=}, data=0x{self.data.hex()}"


class WriteRequestReply(RMAPPacket):
    """An RMAP Reply packet to a Write Request."""

    def __init__(self, data: Union[bytes, np.ndarray]):
        super().__init__(data)
        self._status = data[3]

    @classmethod
    def is_write_reply(cls, data: Union[bytes, np.ndarray]):
        if not RMAPPacket.is_rmap_packet(data):
            return False
        if data[0] != 0x50:
            return False
        if (data[2] == 0x3C or data[2] == 0x2C) and data[4] == 0x51:
            return True

    @property
    def status(self):
        return self._status

    def __str__(self):
        return f"Write Request Reply: status={self.status}"


class ReadRequest(RMAPPacket):
    """A Read Request SpaceWire RMAP Packet."""

    def __init__(self, data: Union[bytes, np.ndarray]):
        super().__init__(data)

    @classmethod
    def is_read_request(cls, data: Union[bytes, np.ndarray]):
        if not RMAPPacket.is_rmap_packet(data):
            return False
        if data[0] != 0x51:
            return False
        if data[2] == 0x4C and data[3] == 0xD1:
            return True
        return False

    @property
    def address(self):
        return get_address(self._bytes)

    @property
    def data_length(self):
        return get_data_length(self._bytes)

    def __str__(self):
        return f"Read Request: tid={self.transaction_id}, address=0x{self.address:04x}, data length={self.data_length}"


class ReadRequestReply(RMAPPacket):
    """An RMAP Reply packet to a Read Request."""

    def __init__(self, data: Union[bytes, np.ndarray]):
        super().__init__(data)

    @classmethod
    def is_read_reply(cls, data: Union[bytes, np.ndarray]):
        if not RMAPPacket.is_rmap_packet(data):
            return False
        if data[0] != 0x50:
            return False
        if data[2] == 0x0C and data[4] == 0x51:
            return True

    @property
    def data(self) -> bytes:
        return get_data(self._bytes)

    @property
    def data_length(self):
        return get_data_length(self._bytes)

    def __str__(self):
        data_length = self.data_length
        return (
            f"Read Request Reply: data length={data_length}, data={self.data[:20]} "
            f"{'(data is cut to max 20 bytes)' if data_length > 20 else ''}\n"
        )


class SpaceWireInterface:
    """
    This interface defines methods that are used by the DPU to communicate with the FEE over
    SpaceWire.
    """

    def __enter__(self):
        self.connect()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def connect(self):
        raise NotImplementedError

    def disconnect(self):
        raise NotImplementedError

    def configure(self):
        raise NotImplementedError

    def flush(self):
        raise NotImplementedError

    def send_timecode(self, timecode: int):
        raise NotImplementedError

    def read_packet(self, timeout: int = None) -> Tuple[int, bytes]:
        """
        Read a full packet from the SpaceWire transport layer.

        Args:
            timeout (int): timeout in milliseconds [default=None]
        Returns:
            A tuple with the terminator value and a bytes object containing the packet.
        """
        raise NotImplementedError

    def write_packet(self, packet: bytes):
        """
        Write a full packet to the SpaceWire transport layer.

        Args:
            packet (bytes): a bytes object containing the SpaceWire packet

        Returns:
            None.
        """
        raise NotImplementedError

    def read_register(self, address: int, length: int = 4, strict: bool = True) -> bytes:
        """
        Reads the data for the given register from the FEE memory map.

        This function sends an RMAP read request for the register to the FEE.

        Args:
            address: the start address (32-bit aligned) in the remote memory
            length: the number of bytes to read from the remote memory [default = 4]
            strict: perform strict checking of address and length

        Returns:
            data: the 32-bit data that was read from the FEE.
        """
        raise NotImplementedError

    def write_register(self, address: int, data: bytes):
        """
        Writes the data from the given register to the N-FEE memory map.

        The function reads the data for the registry from the local register map
        and then sends an RMAP write request for the register to the N-FEE.

        .. note:: it is assumed that the local register map is up-to-date.

        Args:
            address: the start address (32-bit aligned) in the remote memory
            data: the data that will be written into the remote memory

        Raises:
            RMAPError: when data can not be written on the target, i.e. the N-FEE.
        """

        raise NotImplementedError

    def read_memory_map(self, address: int, size: int):
        """
        Read (part of) the memory map from the N-FEE.

        Args:
            address: start address
            size: number of bytes to read

        Returns:
            a bytes object containing the requested memory map.
        """

        raise NotImplementedError


# General RMAP helper functions ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


def rmap_crc_check(data, start, length) -> int:
    """Calculate the checksum for the given data."""
    return crc_calc(data, start, length)


def get_protocol_id(data: bytes) -> int:
    """
    Returns the protocol identifier field. The protocol ID is 1 (0x01) for the RMAP protocol.
    """
    return data[1]


def get_reply_address_field_length(rx_buffer) -> int:
    """Returns the size of reply address field.

    This function returns the actual size of the reply address field. It doesn't return
    the content of the reply address length field. If you need that information, use the
    reply_address_length() function that work on the instruction field.

    Returns:
         length: the size of the reply address field.
    """
    instruction = get_instruction_field(rx_buffer)
    return reply_address_length(instruction) * 4


def get_data(rxbuf) -> bytes:
    """
    Return the data from the RMAP packet.

    Raises:
        ValueError: if there is no data section in the packet (TODO: not yet implemented)
    """
    instruction_field = get_instruction_field(rxbuf)
    address_length = get_reply_address_field_length(rxbuf)
    data_length = get_data_length(rxbuf)

    offset = 12 if is_read(instruction_field) else 16

    return rxbuf[offset + address_length : offset + address_length + data_length]


def check_data_crc(rxbuf):
    instruction_field = get_instruction_field(rxbuf)
    address_length = get_reply_address_field_length(rxbuf)
    data_length = get_data_length(rxbuf)

    offset = 12 if is_read(instruction_field) else 16
    idx = offset + address_length

    d_crc = rxbuf[idx + data_length]
    c_crc = rmap_crc_check(rxbuf, idx, data_length) & 0xFF
    if d_crc != c_crc:
        raise CheckError(
            f"Data CRC doesn't match calculated CRC, d_crc=0x{d_crc:02X} & c_crc=0x{c_crc:02X}", RMAP_GENERAL_ERROR
        )


def check_header_crc(rxbuf):
    instruction_field = get_instruction_field(rxbuf)
    if is_command(instruction_field):
        offset = 15
    elif is_write(instruction_field):
        offset = 7
    else:
        offset = 11

    idx = offset + get_reply_address_field_length(rxbuf)
    h_crc = rxbuf[idx]
    c_crc = rmap_crc_check(rxbuf, 0, idx)
    if h_crc != c_crc:
        raise CheckError(
            f"Header CRC doesn't match calculated CRC, h_crc=0x{h_crc:02X} & c_crc=0x{c_crc:02X}", RMAP_GENERAL_ERROR
        )


def get_data_length(rxbuf) -> int:
    """
    Returns the length of the data in bytes.

    Raises:
        TypeError: when this method is used on a Write Request Reply packet (which has no
            data length).
    """
    instruction_field = get_instruction_field(rxbuf)

    if not is_command(instruction_field) and is_write(instruction_field):
        raise TypeError(
            "There is no data length field for Write Request Reply packets, "
            "asking for the data length is an invalid operation."
        )

    offset = 12 if is_command(instruction_field) else 8
    idx = offset + get_reply_address_field_length(rxbuf)

    # We could use two alternative decoding methods here:
    #   int.from_bytes(rxbuf[idx:idx+3], byteorder='big')    (timeit=1.166s)
    #   struct.unpack('>L', b'\x00' + rxbuf[idx:idx+3])[0]   (timeit=0.670s)
    data_length = struct.unpack(">L", b"\x00" + rxbuf[idx : idx + 3])[0]
    return data_length


def get_address(rxbuf) -> int:
    """
    Returns the address field (including the extended address field if the address is 40-bits).

    Raises:
        TypeError: when this method is used on a Reply packet (which has no address field).
    """
    instruction_field = get_instruction_field(rxbuf)

    if not is_command(instruction_field):
        raise TypeError("There is no address field for Reply packets, asking for the address is an invalid operation.")

    idx = 7 + get_reply_address_field_length(rxbuf)
    extended_address = rxbuf[idx]
    idx += 1
    address = struct.unpack(">L", rxbuf[idx : idx + 4])[0]
    if extended_address:
        address = address + (extended_address << 32)
    return address


def get_instruction_field(rxbuf):
    idx = 2
    return rxbuf[idx]


def get_transaction_identifier(rxbuf):
    idx = 5 + get_reply_address_field_length(rxbuf)
    tid = struct.unpack(">h", rxbuf[idx : idx + 2])[0]
    return tid


# Functions to interpret the Instrument Field


def is_reserved(instruction):
    """The reserved bit of the 2-bit packet type field from the instruction field.

    For PLATO this bit shall be zero as the 0b10 and 0b11 packet field values are reserved.

    Returns:
        bit value: 1 or 0.
    """
    return (instruction & 0b10000000) >> 7


def is_command(instruction):
    """Returns True if the RMAP packet is a command packet."""
    return (instruction & 0b01000000) >> 6


def is_reply(instruction):
    """Returns True if the RMAP packet is a reply to a previous command packet."""
    return not is_command(instruction)


def is_write(instruction):
    """Returns True if the RMAP packet is a write request command packet."""
    return (instruction & 0b00100000) >> 5


def is_read(instruction):
    """Returns True if the RMAP packet is a read request command packet."""
    return not is_write(instruction)


def is_verify(instruction):
    """Returns True if the RMAP packet needs to do a verify before write."""
    return (instruction & 0b00010000) >> 4


def is_reply_required(instruction):
    """Returns True if the reply bit is set in the instruction field.

    Args:
        instruction (int): the instruction field of an RMAP packet

    .. note:: the name of this function might be confusing.

        This function does **not** test if the packet is a reply packet, but it checks
        if the command requests a reply from the target. If you need to test if the
        packet is a command or a reply, use the is_command() or is_reply() function.

    """
    return (instruction & 0b00001000) >> 3


def is_increment(instruction):
    """Returns True if the data is written to sequential memory addresses."""
    return (instruction & 0b00000100) >> 2


def reply_address_length(instruction):
    """Returns the content of the reply address length field.

    The size of the reply address field is then decoded from the following table:

        Address Field Length  |  Size of Address Field
        ----------------------+-----------------------
             0b00             |      0 bytes
             0b01             |      4 bytes
             0b10             |      8 bytes
             0b11             |     12 bytes

    """
    return (instruction & 0b00000011) << 2



import contextlib
import logging
import re
import subprocess
import sys
import textwrap
from pathlib import Path
from typing import Annotated

import rich
import typer
from rich.console import Console
from rich.table import Table

from cgse_common import AppState
from egse.config import find_files
from egse.env import get_conf_data_location
from egse.env import get_site_id
from egse.plugin import entry_points
from egse.process import ProcessInfo
from egse.setup import Setup
from egse.system import format_datetime

app = typer.Typer()


@app.command()
def top(ctx: typer.Context):
    """
    A top-like interface for core services and device control servers.

    Not yet implemented.
    """
    state: AppState = ctx.obj

    print(f"The `verbose` flag is {state.verbose}")
    print("This fancy top is not yet implemented.")


@app.command()
def init(project: str = ""):
    """Initialize your project."""
    from rich.prompt import Prompt, Confirm

    project = project.upper()
    site_id = None

    rich.print(r"[light_steel_blue]Please note default values are given between \[brackets].[/]")

    answer = Prompt.ask(f"What is the name of the project [{project}] ?")
    if answer:
        project = answer.upper()
    while not site_id:
        answer = Prompt.ask("What is the site identifier ?")
        if answer:
            site_id = answer.upper()
        else:
            answer = Confirm.ask("Abort?")
            if answer:
                return

    data_storage_location = f"~/data/{project}/{site_id}/"
    answer = Prompt.ask(f"Where can the project data be stored [{data_storage_location}] ?")
    if answer:
        data_storage_location = answer

    conf_data_location = f"~/data/{project}/{site_id}/conf/"
    answer = Prompt.ask(f"Where will the configuration data be located [{conf_data_location}] ?")
    if answer:
        conf_data_location = answer

    log_file_location = f"~/data/{project}/{site_id}/log/"
    answer = Prompt.ask(f"Where will the logging messages be stored [{log_file_location}] ?")
    if answer:
        log_file_location = answer

    local_settings_path = f"~/data/{project}/{site_id}/local_settings.yaml"
    answer = Prompt.ask(f"Where shall I create a local settings YAML file [{local_settings_path}] ?")
    if answer:
        local_settings_path = answer

    Path(data_storage_location).expanduser().mkdir(exist_ok=True, parents=True)
    (Path(data_storage_location).expanduser() / "daily").mkdir(exist_ok=True, parents=True)
    (Path(data_storage_location).expanduser() / "obs").mkdir(exist_ok=True, parents=True)
    Path(conf_data_location).expanduser().mkdir(exist_ok=True, parents=True)
    Path(log_file_location).expanduser().mkdir(exist_ok=True, parents=True)

    if not Path(local_settings_path).expanduser().exists():
        with open(Path(local_settings_path).expanduser(), "w") as fd:
            fd.write(
                textwrap.dedent(
                    f"""\
                    SITE:
                        ID:  {site_id}
                    """
                )
            )

    answer = Confirm.ask("Shall I add the environment to your ~/bash_profile ?")
    if answer:
        with open(Path("~/.bash_profile").expanduser(), "a") as fd:
            fd.write(
                textwrap.dedent(
                    f"""
                    # Environment for project {project} added by `cgse init` at {format_datetime()}
                    export PROJECT={project}
                    export SITE_ID={site_id}
                    export {project}_DATA_STORAGE_LOCATION={data_storage_location}
                    export {project}_CONF_DATA_LOCATION={conf_data_location}
                    export {project}_LOG_FILE_LOCATION={log_file_location}
                    export {project}_LOCAL_SETTINGS={local_settings_path}
                    """
                )
            )
    else:
        rich.print(
            textwrap.dedent(
                f"""
                # -> Add the following lines to your bash profile or equivalent
                
                export PROJECT={project}
                export SITE_ID={site_id}
                export {project}_DATA_STORAGE_LOCATION={data_storage_location}
                export {project}_CONF_DATA_LOCATION={conf_data_location}
                export {project}_LOG_FILE_LOCATION={log_file_location}
                export {project}_LOCAL_SETTINGS={local_settings_path}
            """
            )
        )


show = typer.Typer(help="Show information about settings, environment, setup, ...")


@show.command(name="settings")
def show_settings():
    """Show the settings that are defined by the installed packages."""
    proc = subprocess.Popen([sys.executable, "-m", "egse.settings"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    rich.print(stdout.decode(), end="")
    if stderr:
        rich.print(f"[red]{stderr.decode()}[/]")


@show.command(name="env")
def show_env(
    mkdir: Annotated[bool, typer.Option(help="Create the missing folder")] = None,
    full: Annotated[bool, typer.Option(help="Provide additional info")] = None,
    doc: Annotated[bool, typer.Option(help="Provide documentation on environment variables")] = None,
):
    """Show the environment variables that are defined for the project."""
    options = [opt for opt, flag in [("--mkdir", mkdir), ("--full", full), ("--doc", doc)] if flag]

    cmd = [sys.executable, "-m", "egse.env"]
    cmd += options if options else []

    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    rich.print(stdout.decode(), end="")
    if stderr:
        rich.print(f"[red]{stderr.decode()}[/]")


@show.command(name="procs")
def show_processes():
    """Show the processes that are running for the installed packages."""
    processes: list[ProcessInfo] = []

    for ep in entry_points("cgse.explore"):
        # print(f"{ep.name = }, {ep.module = }, {ep.load() = }, {ep.extras = }")
        explore = ep.load()
        with contextlib.suppress(AttributeError):
            show_procs = getattr(explore, "show_processes")
            processes.extend(show_procs())

    console = Console()
    console.print(create_rich_table(processes))


def create_rich_table(processes: list[ProcessInfo]):
    """Create a rich table with the processes from all registered packages."""

    table = Table(title="Process Information")

    table.add_column("PID", style="cyan", no_wrap=True)
    table.add_column("Name", style="magenta")
    table.add_column("User", style="green")
    table.add_column("CPU%", justify="right", style="yellow")
    # table.add_column("Memory%", justify="right", style="blue")
    table.add_column("Memory MB", justify="right", style="blue")
    table.add_column("Status", style="red")
    table.add_column("Command", style="white")

    # Sort by CPU usage
    processes.sort(key=lambda x: x.cpu_percent, reverse=True)

    for proc in processes:
        proc = proc.as_dict()
        cmd = re.sub(r"^.*/?python(?:\d+(?:\.\d+)?)?\s+(?:-m)?\s*", "python -m ", proc["command"])
        table.add_row(
            str(proc["pid"]),
            proc["name"],
            proc.get("username", "N/A"),
            f"{proc.get('cpu_percent', 0):.1f}",
            # f"{proc.get('memory_percent', 0):.1f}",
            f"{proc.get('memory_mb', 0):.1f}",
            proc.get("status", "unknown"),
            cmd[:100] + "..." if len(cmd) > 100 else cmd,
        )

    return table


@show.command(name="setup")
def show_setup(
    use_cm: Annotated[bool, typer.Option(help="Get the current setup from the configuration manager")] = True,
    list_all: Annotated[bool, typer.Option(help="Print a list of all available setups")] = False,
    setup_id: Annotated[int, typer.Option(help="Show the setup with this id")] = -1,
):
    if use_cm:
        try:
            from egse.confman import ConfigurationManagerProxy
        except ImportError:
            rich.print("[dark_orange3]WARNING: package 'cgse-core' is not installed, service not available.[/]")
            return

        with ConfigurationManagerProxy() as cm:
            if list_all:
                rich.print(cm.list_setups())
            else:
                rich.print(cm.get_setup())
        return

    location = get_conf_data_location()
    site_id = get_site_id()

    if list_all:
        files = find_files(f"SETUP_{site_id}_*_*.yaml", root=location)
        files = list(files)
        if files:
            location = files[0].parent.resolve()
        rich.print(sorted([f.name for f in files]))
        rich.print(f"Loaded from [purple]{location}.")
    else:
        if setup_id == -1:
            setup_files = find_files(f"SETUP_{site_id}_*_*.yaml", root=location)
        else:
            setup_files = find_files(f"SETUP_{site_id}_{setup_id:05d}_*.yaml", root=location)
        setup_files = list(setup_files)
        if len(setup_files) > 0:
            setup_file = sorted(setup_files)[-1]
            setup = Setup.from_yaml_file(setup_file)
            rich.print(setup)
        else:
            rich.print("[red]No setup files were found.[/]")


check = typer.Typer(help="Check installation, settings, required files, etc.")


@check.command(name="setups")
def check_setups(verbose: bool = False):
    """Perform a number of checks on the SETUP files."""

    logging.basicConfig(
        level=logging.CRITICAL,
        format="[%(asctime)s] %(threadName)-12s %(levelname)-8s %(name)-20s %(lineno)5d:%(module)-20s %(message)s",
    )

    # What can we check with respect to the setups?
    #
    # - CONF_DATA_LOCATION

    from egse.env import get_conf_data_location
    from egse.env import get_site_id
    from egse.config import find_files

    any_errors = 0
    error_messages = []

    try:
        conf_data_location = get_conf_data_location()
    except ValueError:
        conf_data_location = None

    site_id = get_site_id()

    # ---------- check if the Site_ID is set

    verbose and rich.print("Checking site id ...")

    if not site_id:
        any_errors += 1
        error_messages.append(
            "[red]The environment variable SITE_ID is not set. "
            "SITE_ID is required to define the project settings and environment variables.[/]"
        )

    # ---------- check if the <PROJECT>_CONF_DATA_LOCATION is set

    verbose and rich.print("Checking configuration data location ...")

    if not conf_data_location:
        any_errors += 1
        error_messages.append(
            "[red]The location of the configuration data can not be determined, check your environment using `cgse "
            "show env`.[/]"
        )
    elif not Path(conf_data_location).exists():
        any_errors += 1
        error_messages.append(f"[red]The location of the configuration data doesn't exist: {conf_data_location!s}[/]")

    if any_errors:
        print_error_messages(error_messages)
        return

    # ---------- check if there is at least one SETUP in the configuration data folder

    verbose and rich.print("Checking available SETUP files ...")

    files = list(find_files("SETUP*.yaml", root=conf_data_location))

    if not files:
        any_errors += 1
        error_messages.append(f"[red]No SETUP files were found at {conf_data_location}[/]")
    else:
        verbose and rich.print(f":arrow_forward: Found {len(files)} Setup files.")

    regex = re.compile(f"SETUP_{site_id}_00000_.*.yaml")

    if not any(True for file in files if regex.search(str(file))):
        any_errors += 1
        error_messages.append(f"[red]There is no 'Zero' SETUP for {site_id} in {conf_data_location}[/]")
    else:
        verbose and rich.print(f":arrow_forward: Found the 'Zero' Setup file: SETUP_{site_id}_00000_*.yaml")

    if any_errors:
        print_error_messages(error_messages)
        return

    # ---------- check for each SETUP file if the site_id matches the SITE_ID

    verbose and rich.print("Checking site_id in setup files ...")

    for file in files:
        setup = Setup.from_yaml_file(file)
        this_site_id = setup.get("site_id")
        if this_site_id is None:
            any_errors += 1
            error_messages.append(f"There is no 'site_id' defined in '{file!s}'")
        elif this_site_id != site_id:
            any_errors = 1
            error_messages.append(
                f"[red]The site_id ('{this_site_id}') in '{file!s}' doesn't match the environment "
                f"variable SITE_ID={site_id}[/]"
            )

    if any_errors:
        print_error_messages(error_messages)
    else:
        rich.print("[green]everything seems to be ok.[/]")


def print_error_messages(messages: list[str]):
    from rich.panel import Panel

    content = "\n"

    for msg in messages:
        content += f":arrow_forward: {msg}\n"
    content += "\n[orange1]Fix the above errors before proceeding.[/]"

    rich.print(Panel(content, title="Errors found during analysis"))



"""
This module provides the services to the control servers.

Each control server has a services protocol which provides commands that will
be executed on the control server instead of the device controller. This is
typically used to access control server specific settings like monitoring frequency,
logging levels, or to quit the control server in a controlled way.

"""

import inspect
import logging
import textwrap
from pathlib import Path

from egse.command import ClientServerCommand
from egse.control import ControlServer
from egse.decorators import dynamic_interface
from egse.log import logger
from egse.protocol import CommandProtocol
from egse.proxy import Proxy
from egse.proxy import REQUEST_TIMEOUT
from egse.registry.client import RegistryClient
from egse.settings import Settings
from egse.zmq_ser import bind_address
from egse.zmq_ser import connect_address

HERE = Path(__file__).parent

SERVICE_SETTINGS = Settings.load(location=HERE, filename="services.yaml")


class ServiceCommand(ClientServerCommand):
    pass


class ServiceProtocol(CommandProtocol):
    def __init__(self, control_server: ControlServer):
        super().__init__(control_server)

        self.load_commands(SERVICE_SETTINGS.Commands, ServiceCommand, ServiceProtocol)

    def get_bind_address(self):
        """
        Returns a string with the bind address, the endpoint, for accepting connections
        and bind a socket to. The port to connect to is the service port in this case,
        not the commanding port.

        Returns:
            a string with the protocol and port to bind a socket to.
        """
        return bind_address(
            self._control_server.get_communication_protocol(),
            self._control_server.get_service_port(),
        )

    def handle_set_monitoring_frequency(self, freq: float):
        """
        Sets the monitoring frequency (Hz) to the given freq value. This is only approximate since the frequency is
        converted into a delay time and the actual execution of the status function is subject to the load on the
        server and the overhead of the timing.

        Args:
            freq: frequency of execution (Hz)

        Returns:
            Sends back the selected delay time in milliseconds.
        """
        delay = self.get_control_server().set_mon_delay(1.0 / freq)

        logger.debug(f"Set monitoring frequency to {freq}Hz, ± every {delay:.0f}ms.")

        self.send(delay)

    def handle_set_hk_frequency(self, freq: float):
        """
        Sets the housekeeping frequency (Hz) to the given freq value. This is only approximate since the frequency is
        converted into a delay time and the actual execution of the `housekeeping` function is subject to the load on
        the server and the overhead of the timing.

        Args:
            freq: frequency of execution (Hz)

        Returns:
            Sends back the selected delay time in milliseconds.
        """
        delay = self.get_control_server().set_hk_delay(1.0 / freq)

        logger.debug(f"Set housekeeping frequency to {freq}Hz, ± every {delay:.0f}ms.")

        self.send(delay)

    def handle_set_logging_level(self, *args, **kwargs):
        """
        Set the logging level for the logger with the given name.

        When 'all' is given for the name of the logger, the level of all loggers for which the name
        starts with 'egse' will be changed to `level`.

        Args:
            name (str): the name of an existing Logger
            level (int): the logging level

        Returns:
            Sends back an info message on what level was set.
        """
        if args:
            name = args[0]
            level = int(args[1])
        else:
            name = kwargs["name"]
            level = int(kwargs["level"])

        if name == "all":
            for logger in [
                logging.getLogger(logger_name)
                for logger_name in logging.root.manager.loggerDict
                if logger_name.startswith("egse")
            ]:
                logger.setLevel(level)
            msg = f"Logging level set to {level} for ALL 'egse' loggers"
        elif name in logging.root.manager.loggerDict:
            logger = logging.getLogger(name)
            logger.setLevel(level)
            msg = f"Logging level for {name} set to {level}."
        else:
            msg = f"Logger with name '{name}' doesn't exist at the server side."

        # self.control_server.set_logging_level(level)
        logging.debug(msg)
        self.send(msg)

    def handle_quit(self):
        logger.info(f"Sending interrupt to {self.control_server.__class__.__name__}.")
        self.control_server.quit()
        self.send(f"Sent interrupt to {self.control_server.__class__.__name__}.")

    def handle_get_process_status(self):
        logger.debug(f"Asking for process status of {self.control_server.__class__.__name__}.")
        self.send(self.get_status())

    def handle_get_cs_module(self):
        """
        Returns the module in which the control server has been implemented.
        """
        logger.debug(f"Asking for module of {self.control_server.__class__.__name__}.")
        self.send(inspect.getmodule(self.control_server).__spec__.name)

    def handle_get_average_execution_times(self):
        logger.debug(f"Asking for average execution times of {self.control_server.__class__.__name__} functions.")
        self.send(self.control_server.get_average_execution_times())

    def handle_get_storage_mnemonic(self):
        logger.debug(f"Asking for the storage menmonic of {self.control_server.__class__.__name__}.")
        self.send(self.control_server.get_storage_mnemonic())

    def handle_add_listener(self, listener: dict):
        logger.debug(f"Add listener to {self.control_server.__class__.__name__}: {listener}")
        try:
            self.control_server.listeners.add_listener(listener)
            logger.info(f"Registered listener: {listener['name']} with proxy {listener['proxy']}")
            self.send(("ACK",))
        except ValueError as exc:
            self.send(("NACK", exc))  # Why not send back a failure object?

    def handle_remove_listener(self, listener: dict):
        logger.debug(f"Remove listener from {self.control_server.__class__.__name__}: {listener}")
        try:
            self.control_server.listeners.remove_listener(listener)
            logger.info(f"Removed listener: {listener['name']}")
            self.send(("ACK",))
        except ValueError as exc:
            self.send(("NACK", exc))  # Why not send back a failure object?

    def handle_get_listener_names(self):
        logger.debug(f"Get names of registered listener from {self.control_server.__class__.__name__}")
        try:
            names = self.control_server.listeners.get_listener_names()
            self.send((names,))
        except ValueError as exc:
            self.send(("", exc))  # Why not sent back a Failure object?

    def handle_register_to_storage(self):
        logger.debug("(re-)registering to the storage manager")
        try:
            self.control_server.register_to_storage_manager()
            self.send(("ACK",))
        except Exception as exc:
            self.send(("NACK", exc))  # Why not send back a failure object?


class ServiceInterface:
    @dynamic_interface
    def set_monitoring_frequency(self, freq: float): ...
    @dynamic_interface
    def set_hk_frequency(self, freq: float): ...
    @dynamic_interface
    def set_logging_level(self, name: str, level: int): ...
    @dynamic_interface
    def quit_server(self): ...
    @dynamic_interface
    def get_process_status(self): ...
    @dynamic_interface
    def get_cs_module(self): ...
    @dynamic_interface
    def get_average_execution_times(self): ...
    @dynamic_interface
    def get_storage_mnemonic(self): ...
    @dynamic_interface
    def add_listener(self, listener: dict): ...
    @dynamic_interface
    def remove_listener(self, listener: dict): ...
    @dynamic_interface
    def get_listener_names(self, listener: dict): ...
    @dynamic_interface
    def register_to_storage(self): ...


class ServiceProxy(Proxy, ServiceInterface):
    """
    A ServiceProxy is a simple class that forwards service commands to a control server.
    """

    def __init__(
        self,
        service_type: str = None,
        protocol: str = "tcp",
        hostname: str = None,
        port: int = None,
        timeout: int = REQUEST_TIMEOUT,
    ):
        """
        The Service Proxy class is used to send service commands to the control server.

        Args:
            service_type: the target service type
            protocol: the transport protocol [default: tcp]
            hostname: the IP address of the control server
            port: the port on which the control server is listening for service commands
            timeout: number of milliseconds before a timeout is triggered
        """

        if hostname is None or port is None:
            if service_type is None:
                raise ValueError(
                    textwrap.dedent(
                        """\
                        No service_type or hostname/port information provided.
                        I need either the service_type or the endpoint information to create a ServiceProxy.
                        """
                    )
                )

            with RegistryClient() as reg:
                service = reg.discover_service(service_type)

                if service:
                    protocol = service.get("protocol", "tcp")
                    hostname = service["host"]
                    port = service["metadata"]["service_port"]

                else:
                    raise RuntimeError(f"No service registered as {service_type}")

        endpoint = connect_address(protocol, hostname, port)
        super().__init__(endpoint, timeout)



import logging

from egse.decorators import dynamic_interface


_LOGGER = logging.getLogger(__name__)


class AlphaControllerInterface:
    @dynamic_interface
    def info(self):
        """Returns basic information about the hexapod and the controller.

        Returns:
            a multiline response message containing the device info.
        Raises:
            HexapodError: when information can not be retrieved.
        """
        raise NotImplementedError

    @dynamic_interface
    def reset(self, wait=True):
        """Completely resets the Hexapod hardware controller with the standard boot cycle.

        Args:
            wait (bool): after the reset command has been sent to the controller, wait
                for 30 seconds which should complete the cycle, i.e. this command will
                only return after 30 seconds.

        Note:
           This command is equivalent to power cycling the controller manually.

        """
        raise NotImplementedError

    @dynamic_interface
    def homing(self):
        """Start the homing cycle for the Hexapod PUNA.

        Homing is required before performing a control movement. Without absolute encoders,
        the homing is performed with a hexapod movement until detecting the reference sensor
        on each of the actuators. The Hexapod will go to a position were the sensors are
        reached that signal a known calibrated position and then returns to the zero position.

        Whenever a homing is performed, the method will return before the actual movement
        is finished.

        The homing cycle takes about two minutes to complete, but the ``homing()`` method
        returns almost immediately. Therefore, to check if the homing is finished, use
        the is_homing_done() method.

        Returns:
            0 on success.
        Raises:
            HexapodError: when there is a timeout or when there is a communication error with
            the hexapod.
        """
        raise NotImplementedError

    @dynamic_interface
    def set_virtual_homing(self, tx, ty, tz, rx, ry, rz):
        """Starts the virtual homing cycle on the hexapod.

        This command uses the position given in parameters to initialize the hexapod position.
        No movements of the hexapod are performed during this homing cycle. Please note that the
        position specified in parameters must match the absolute position of the Object coordinate
        system in the User coordinate system (see description in the manual chapter 2 on coordinates
        systems). This position correspond to the answer of the command `get_user_positions()`.
        During this operation, it is important to have the same hexapod position as those defined
        during the record of the position. Otherwise, the system initialization will be incorrect.

        Args:
            tx (float): position on the X-axis [mm]
            ty (float): position on the Y-axis [mm]
            tz (float): position on the Z-axis [mm]
            rx (float): rotation around the X-axis [deg]
            ry (float): rotation around the Y-axis [deg]
            rz (float): rotation around the Z-axis [deg]

        Returns:
            return_code: 0 on success, -1 when ignored

        Raises:
            HexapodError: when the arguments do not match up, or when there is a timeout or when
            there is a
            socket communication error.

        """
        return NotImplementedError

    @dynamic_interface
    def stop(self):
        """Stop the current motion. This command can be sent during a motion of the Hexapod.

        Returns:
            0 on success.
        Raises:
            HexapodError: when there is a timeout or when there is a communication error with
            the hexapod.
        """
        raise NotImplementedError

    @dynamic_interface
    def clear_error(self):
        """Clear all errors in the controller software.

        Returns:
            0 on success.
        Raises:
            HexapodError: when there is Time-Out or when there is a communication error with the
            hexapod.
        """
        raise NotImplementedError

    @dynamic_interface
    def activate_control_loop(self):
        """Activates the control loop on motors.

        Returns:
            0 on success, -1 when ignored, -2 on error.
        Raises:
            HexapodError: when there is a timeout or when there is a communication error with
            the hexapod
                hardware controller.
        """
        raise NotImplementedError

    @dynamic_interface
    def deactivate_control_loop(self):
        """Disables the control loop on the servo motors.

        Returns:
            0 on success.
        Raises:
            HexapodError: when there is a timeout or when there is a communication error with
            the hexapod.
        """
        raise NotImplementedError

    @dynamic_interface
    def jog(self, axis: int, inc: float) -> int:
        """Perform a JOG-type movement on the specified actuator.

        Note:
            This is a maintenance feature.

        Args:
            axis (int): number of the actuator (1 to 6)
            inc (float): increment to achieve in mm
        Returns:
            0 on success, -1 if command was ignored due to non-compliance.
        Raises:
            HexapodError: when there is a timeout or when there is a communication error with
            the hexapod.
        """
        raise NotImplementedError

    @dynamic_interface
    def get_debug_info(self):
        """
        Request debug information from the Hexapod Controller.

        The method returns four values that represent the following status:

        1. ``Homing``: state of the homing
        2. ``PosRef``: state of the Position Reference command
        3. ``KinError``: error in kinematic calculation
        4. ``Panel``: Panel state

        The Homing can take the following values:

            =====     ==================
            Value     Meaning
            =====     ==================
            0         Undefined
            1         Homing in progress
            2         Homing done
            3         Error
            =====     ==================

        The PosRef can take the following values:

            =======     =====================
             Value       Meaning
            =======     =====================
               0         Undefined
               1         Abort input error
               2         Movement in progress
               3         Position reached
               4         Error
            =======     =====================

        The KinError can take the following values:

            =======     ===============================================
             Value       Meaning
            =======     ===============================================
               0         none
               1         Homing not done
               2         Inverse kinematic model (MGI) – workspace
               3         Inverse kinematic model (MGI) – joint angle
               4         Forward kinematic model (MGD) – workspace
               5         Forward kinematic model (MGD) – max iteration
               6         Position calculation (PLCC) – workspace
               7         Position calculation (PLCC) – max iteration.
            =======     ===============================================

        The Panel status can take the following values:

            ======   ===============
            Index       Command
            ======   ===============
            -2       Command error
            -1       Ignored / Command not allowed
            **0**       **None**
            1        Homing
            2        Stop
            3        Control ON
            4        Control OFF
            10       Valid POS
            11       Move
            12       Sequence
            13       Specific POS
            15       Clear Error
              **Family “Set config”**
            ------------------------
            21       Config CS
            22       Config Limits_mTn
            23       Config Limits_uTo
            24       Config Limits_Enabled
            25       Config Speed
            26       Config Current
            27       Config Backlash
              **Family “Get config”**
            ------------------------
            31       Config CS
            32       Config Limits_mTn
            33       Config Limits_uTo
            34       Config Limits_Enabled
            35       Config Speed
            36       Config Current
            37       Config Backlash
              **Family “Maintenance”**
            ------------------------
            41       Jog
            50       Config Save
            51       Config Default
            52       Config Saved?
            55       Version
            ======   ===============

        """
        raise NotImplementedError

    @dynamic_interface
    def configure_coordinates_systems(self, tx_u, ty_u, tz_u, rx_u, ry_u, rz_u, tx_o, ty_o, tz_o, rx_o, ry_o, rz_o):
        """
        Change the definition of the User Coordinate System and the Object Coordinate System.

        The parameters tx_u, ty_u, tz_u, rx_u, ry_u, rz_u are used to define the user coordinate
        system
        relative to the Machine Coordinate System and the parameters tx_o, ty_o, tz_o, rx_o,
        ry_o, rz_o
        are used to define the Object Coordinate System relative to the Platform Coordinate System.

        Args:
            tx_u (float): translation parameter that define the user coordinate system relative
            to the machine coordinate system [in mm]
            ty_u (float): translation parameter that define the user coordinate system relative
            to the machine coordinate system [in mm]
            tz_u (float): translation parameter that define the user coordinate system relative
            to the  machine coordinate system [in mm]
            rx_u (float): rotation parameter that define the user coordinate system relative to
            the machine coordinate system [in deg]
            ry_u (float): rotation parameter that define the user coordinate system relative to
            the machine coordinate system [in deg]
            rz_u (float): rotation parameter that define the user coordinate system relative to
            the machine coordinate system [in deg]
            tx_o (float): translation parameter that define the object coordinate system relative
            to the platform coordinate system [in mm]
            ty_o (float): translation parameter that define the object coordinate system relative
            to the platform coordinate system [in mm]
            tz_o (float): translation parameter that define the object coordinate system relative
            to the platform coordinate system [in mm]
            rx_o (float): rotation parameter that define the object coordinate system relative to the platform
            coordinate system [in deg]
            ry_o (float): rotation parameter that define the object coordinate system relative to the platform
            coordinate system [in deg]
            rz_o (float): rotation parameter that define the object coordinate system relative to the platform
            coordinate system [in deg]

        Returns:
            0 on success and -1 when the configuration is ignored, e.g. when password protection
            is enabled.
        """
        raise NotImplementedError

    @dynamic_interface
    def get_coordinates_systems(self):
        """Retrieve the definition of the User Coordinate System and the Object Coordinate System.

        Returns:
            tx_u, ty_u, tz_u, rx_u, ry_u, rz_u, tx_o, ty_o, tz_o, rx_o, ry_o, rz_o where the
            translation \
            parameters are in [mm] and the rotation parameters are in [deg].
        Raises:
            HexapodError: when an error occurred while trying to retrieve the information.
        """
        raise NotImplementedError

    @dynamic_interface
    def get_actuator_length(self):
        """Retrieve the current length of the hexapod actuators.

        Returns:
            array: an array of six float values for actuator length L1 to L6 in [mm], and \
            None: when an Exception was raised and logs the error message.
        """
        raise NotImplementedError

    @dynamic_interface
    def move_absolute(self, tx, ty, tz, rx, ry, rz):
        """Move/define the Object Coordinate System position and orientation expressed
        in the invariant user coordinate system.

        The rotation centre coincides with the Object Coordinates System origin and
        the movements are controlled with translation components at first (Tx, Ty, tZ)
        and then the rotation components (Rx, Ry, Rz).

        Args:
            tx (float): position on the X-axis [mm]
            ty (float): position on the Y-axis [mm]
            tz (float): position on the Z-axis [mm]
            rx (float): rotation around the X-axis [deg]
            ry (float): rotation around the Y-axis [deg]
            rz (float): rotation around the Z-axis [deg]

        Returns:
            return_code: 0 on success, -1 when ignored, -2 on error

        Raises:
            HexapodError: when the arguments do not match up, or when there is a time out or when
            there is a
            socket communication error.

        .. note::
           When the command was not successful, this method will query the
           POSVALID? using the checkAbsolutePosition() and print a summary
           of the error messages to the log file.

        .. todo::
           do we want to check if the movement is valid before actually performing
           the movement?

        """
        raise NotImplementedError

    @dynamic_interface
    def move_relative_object(self, tx, ty, tz, rx, ry, rz):
        """Move the object relative to its current object position and orientation.

        The relative movement is expressed in the object coordinate system.

        Args:
            tx (float): position on the X-axis [mm]
            ty (float): position on the Y-axis [mm]
            tz (float): position on the Z-axis [mm]
            rx (float): rotation around the X-axis [deg]
            ry (float): rotation around the Y-axis [deg]
            rz (float): rotation around the Z-axis [deg]

        Returns:
            0 on success, -1 when ignored, -2 on error.

        Raises:
            HexapodError: when the arguments do not match up, or when there is a time out or when
            there is a
            socket communication error.

        .. todo:: do we want to check if the movement is valid before actually performing
                  the movement?

        """
        raise NotImplementedError

    @dynamic_interface
    def move_relative_user(self, tx, ty, tz, rx, ry, rz):
        """Move the object relative to its current object position and orientation.

        The relative movement is expressed in the (invariant) user coordinate system.

        Args:
            tx (float): position on the X-axis [mm]
            ty (float): position on the Y-axis [mm]
            tz (float): position on the Z-axis [mm]
            rx (float): rotation around the X-axis [deg]
            ry (float): rotation around the Y-axis [deg]
            rz (float): rotation around the Z-axis [deg]

        Returns:
            0 on success, -1 when ignored, -2 on error.

        Raises:
            HexapodError: when the arguments do not match up, or when there is a time out or when
            there is a
            socket communication error.

        .. todo:: do we want to check if the movement is valid before actually performing
                  the movement?

        """
        raise NotImplementedError

    @dynamic_interface
    def check_absolute_movement(self, tx, ty, tz, rx, ry, rz):
        """Check if the requested object movement is valid.

        The absolute movement is expressed in the user coordinate system.

        Args:
            tx (float): position on the X-axis [mm]
            ty (float): position on the Y-axis [mm]
            tz (float): position on the Z-axis [mm]
            rx (float): rotation around the X-axis [deg]
            ry (float): rotation around the Y-axis [deg]
            rz (float): rotation around the Z-axis [deg]

        Returns:
            tuple: where the first element is an integer that represents the
                bitfield encoding the errors. The second element is a dictionary
                with the bit numbers that were (on) and the corresponding error
                description.

        .. todo:: either provide a more user friendly return value or a method/function
                  to process the information into a human readable form. Also update
                  the documentation of this method with more information about the
                  bitfields etc.
        """
        raise NotImplementedError

    @dynamic_interface
    def check_relative_object_movement(self, tx, ty, tz, rx, ry, rz):
        """Check if the requested object movement is valid.

        The relative motion is expressed in the object coordinate system.

        Args:
            tx (float): position on the X-axis [mm]
            ty (float): position on the Y-axis [mm]
            tz (float): position on the Z-axis [mm]
            rx (float): rotation around the X-axis [deg]
            ry (float): rotation around the Y-axis [deg]
            rz (float): rotation around the Z-axis [deg]

        Returns:
            tuple: where the first element is an integer that represents the
                bitfield encoding the errors. The second element is a dictionary
                with the bit numbers that were (on) and the corresponding error
                description.

        .. todo:: either provide a more user friendly return value or a method/function
                  to process the information into a human readable form. Also update
                  the documentation of this method with more information about the
                  bitfields etc.
        """
        raise NotImplementedError

    @dynamic_interface
    def check_relative_user_movement(self, tx, ty, tz, rx, ry, rz):
        """Check if the requested object movement is valid.

        The relative motion is expressed in the user coordinate system.

        Args:
            tx (float): position on the X-axis [mm]
            ty (float): position on the Y-axis [mm]
            tz (float): position on the Z-axis [mm]
            rx (float): rotation around the X-axis [deg]
            ry (float): rotation around the Y-axis [deg]
            rz (float): rotation around the Z-axis [deg]

        Returns:
            tuple: where the first element is an integer that represents the
                bitfield encoding the errors. The second element is a dictionary
                with the bit numbers that were (on) and the corresponding error
                description.

        .. todo:: either provide a more user friendly return value or a method/function
                  to process the information into a human readable form. Also update
                  the documentation of this method with more information about the
                  bitfields etc.
        """
        raise NotImplementedError

    @dynamic_interface
    def get_user_positions(self):
        """Retrieve the current position of the hexapod.

        The returned position corresponds to the position of the Object Coordinate System
        in the User Coordinate System.

        Returns:
            array: an array of six float values for Tx, Ty, Tz, Rx, Ry, Rz.
            None: when an Exception was raised and logs the error message.

        Note: This is equivalent to the POSUSER? command.
        """
        raise NotImplementedError

    @dynamic_interface
    def get_machine_positions(self):
        """Retrieve the current position of the hexapod.

        The returned position corresponds to the position of the Platform Coordinate System
        in the Machine Coordinate System.

        Returns:
            array: an array of six float values for Tx, Ty, Tz, Rx, Ry, Rz.
            None: when a PMACError was raised and logs the error message.

        Note: This is equivalent to the POSMACH? command.
        """
        raise NotImplementedError

    @dynamic_interface
    def set_speed(self, vt, vr):
        """Set the speed of the hexapod movements.

        Args:
            vt (float): The translation speed, expressed in mm per second [mm/s].
            vr (float): The angular speed, expressed in deg per second [deg/s].

        The arguments `vt` and `vr` are automatically limited by the controller
        between the minimum and maximum allowed speeds for the hexapod.
        See the `getSpeed()` method to know the limits.

        Returns:
            0 on success and -1 when the configuration is ignored, e.g. when password protection
            is enabled.
        """
        raise NotImplementedError

    @dynamic_interface
    def get_speed(self):
        """Retrieve the configuration of the movement speed.

        Returns:
            vt, vr, vt_min, vr_min, vt_max, vr_max

        Where:
            * ``vt`` is the translation speed of the hexapod in mm per second [mm/s]
            * ``vr`` is the angular speed of the hexapod in deg per second [deg/s]
            * ``vt_min``, ``vt_max`` are the limits for the translation speed [mm/s]
            * ``vr_min``, ``vr_max`` are the limits for the angular speed [mm/s]

        """
        raise NotImplementedError

    @dynamic_interface
    def get_general_state(self):
        """Retrieve general state information of the hexapod.

        Returns:
            tuple: where the first element is an integer where the bits represent each state, and
                the second element is an array of True/False flags for each state described in
                Hexapod
                Controller API, section 4.5.6.

            None: when an Exception was raised and logs the error message.

        Note: This is equivalent to the STATE#HEXA? Command.
        """
        raise NotImplementedError

    @dynamic_interface
    def get_actuator_state(self):
        """Retrieve general state information of the actuators.

        For each of the six actuators, an integer value is returned that should be interpreted as a
        bit field containing status bits for that actuator.

            ======   ========================
             Bit      Function
            ======   ========================
              0       In position
              1       Control loop on servo motors active
              2       Homing done
              3       Input "Home Switch"
              4       Input "Positive limit switch"
              5       Input "Negative limit switch"
              6       Brake control output
              7       Following error (warning)
              8       Following error (Fatal)
              9       Actuator out of bounds error
             10       Amplifier error
             11       Encoder error
             12       Phasing error (brushless engine only)
             13-23    Reserved
            ======   ========================

        Returns:
            array: an array of six (6) dictionaries with True/False flags for each actuator state
            as described in
                Hexapod Controller API, section 4.5.5.
            None: when an Exception was raised and logs the error message.

        Note: This is equivalent to the STATE#ACTUATOR? Command.
        """
        raise NotImplementedError

    @dynamic_interface
    def goto_specific_position(self, pos):
        """Ask to go to a specific position.

        * pos=0 Zero position (jog & maintenance only!)
        * pos=1 Zero position
        * pos=2 Retracted position

        Returns:
            0 on success, -1 when ignored.

        Raises:
            HexapodError: when there is Time-Out or when there is a communication error with the
            hexapod controller.
        """
        raise NotImplementedError

    @dynamic_interface
    def goto_retracted_position(self):
        """Ask the hexapod to go to the retracted position.

        Returns:
            0 on success, -1 when ignored.

        Raises:
            HexapodError: when there is Time-Out or when there is a socket communication error.
        """
        raise NotImplementedError

    @dynamic_interface
    def goto_zero_position(self):
        """Ask the hexapod to go to the zero position.

        Returns:
            0 on success, -1 when ignored.

        Raises:
            HexapodError: when there is Time-Out or when there is a socket communication error.
        """
        raise NotImplementedError

    @dynamic_interface
    def is_homing_done(self):
        """
        Check if Homing is done. This method checks the ``Q26`` variable.
        When this variable indicates 'Homing is done' it means the command has properly been
        executed,
        but it doesn't mean the Hexapod is in position. The hexapod might still be moving to its
        zero position.

        Returns:
            True when the homing is done, False otherwise.
        """
        raise NotImplementedError

    @dynamic_interface
    def is_in_position(self):
        """
        Checks if the actuators are in position.

        This method queries the Q36 variable and examines the third bit which is the 'Is
        Position' bit.
        This method does **not** query the actuator state variables Q30 till Q36.

        Returns:
            True when in position, False otherwise.
        """
        raise NotImplementedError

    @dynamic_interface
    def perform_maintenance(self, axis):
        """Perform a maintenance cycle which consists of travelling the full range
        on one axis corresponding to the Hexapod machine limits. The movement is
        also in machine coordinate system.

        The ``axis`` argument can take the following values:

        * 1: movement on the X-axis
        * 2: movement on the Y-axis
        * 3: movement on the Z-axis
        * 4: movement around the X-axis
        * 5: movement around the Y-axis
        * 6: movement around the Z-axis
        * 10: all previous cycles chained together

        Args:
            axis (int): on which axis the full range movement is executed
        Returns:
            0 on success, -1 when ignored for non-compliance.
        Raises:
            HexapodError: when there is Time-Out or when there is a socket communication error.
        """
        raise NotImplementedError

    def log_positions(self):
        """
        Log the current position of the hexapod (level=INFO). The positions correspond to

          * the position of the Object Coordinate System in the User Coordinate System, and
          * the position of the Platform Coordinate System in the Machine Coordinate System.

        """

        pos = self.get_user_positions()
        _LOGGER.info(
            f"Object [in User]     : "
            f"{pos[0]:2.5f}, {pos[1]:2.5f}, {pos[2]:2.5f}, {pos[3]:2.5f}, {pos[4]:2.5f}, "
            f"{pos[5]:2.5f}"
        )

        pos = self.get_machine_positions()
        _LOGGER.info(
            f"Platform [in Machine]: "
            f"{pos[0]:2.5f}, {pos[1]:2.5f}, {pos[2]:2.5f}, {pos[3]:2.5f}, {pos[4]:2.5f}, "
            f"{pos[5]:2.5f}"
        )


class AlphaPlusControllerInterface(AlphaControllerInterface):
    @dynamic_interface
    def get_limits_value(self, lim):
        """Three different and independent operational workspace limits are defined on the controller:
            * Factory limits: are expressed in machine coordinate system limits. Those parameters cannot be modified.
            * Machine coordinate system limits: they are expressed in the Machine coordinate system. It can be used to
            secure the hexapod (and/or object) from its environment.
            * User coordinate system limits: they are expressed in the User coordinate system. It can be used to limits
            the movements of the object mounted on hexapod.

        Remark: operational workspace limits must be understood as limits in terms of amplitude of movement. Those
        limits are defined for each operational axis with a negative and positive value and are used in the validation
        process. Position on each operational axis must be within those two values.

        Args:
            lim (int):
            0 = factory (GET only),
            1 = machine cs limit,
            2 = user cs limit

        """
        raise NotImplementedError

    @dynamic_interface
    def get_limits_state(self):
        """Return workspace limits enable state"""
        raise NotImplementedError

    @dynamic_interface
    def get_temperature(self):
        """Return the 6xPT100 temperature sensor's value in C"""
        raise NotImplementedError

    @dynamic_interface
    def machine_limit_enable(self, state):
        """Enable or disable the machine workspace limit of the hexapod.

        Remark: the factory machine coordinate system limit is always enabled to ensure the safety of the hexapod. It
        cannot be disabled.

            state(int):
            0 = disabled
            1 = enabled

        """

        raise NotImplementedError

    @dynamic_interface
    def machine_limit_set(self, *par):
        """Sets the machine workspace limits of the hexapod. Will raise error if not all the parameters are set (see
        Args definition)

        Remark: operational workspace limits must be understood as limits in terms of amplitude of movement. Those limits
        are defined for each operational axis with a negative and positive value and are used in the validation process.
        Position on each operational axis must be within those two values.

        Args:
            ntx(double): negative position limit in X in mm
            nty(double): negative position limit in Y in mm
            ntz(double): negative position limit in Z in mm
            nrx(double): negative position limit around the axis X in deg
            nry(double): negative position limit around the axis Y in deg
            nrz(double): negative position limit around the axis Y in deg

            ptx(double): positive position limit in X in mm
            pty(double): positive position limit in Y in mm
            ptz(double): positive position limit in Z in mm
            prx(double): positive position limit around the axis X in deg
            pry(double): positive position limit around the axis Y in deg
            prz(double): positive position limit around the axis Y in deg

        """

        raise NotImplementedError

    @dynamic_interface
    def user_limit_enable(self, state):
        """Enable or disable the user workspace limit of the hexapod.

        Remark: the factory machine coordinate system limit is always enabled to ensure the safety of the hexapod. It
        cannot be disabled.

            state(int):
            0 = disabled
            1 = enabled

        """

        raise NotImplementedError

    @dynamic_interface
    def user_limit_set(self, *par):
        """Sets the user workspace limits of the hexapod. Will raise error if not all the parameters are set (see
        Args definition)

        Remark: operational workspace limits must be understood as limits in terms of amplitude of movement. Those
        limits are defined for each operational axis with a negative and positive value and are used in the
        validation process.  Position on each operational axis must be within those two values.

        Args:
            ntx(double): negative position limit in X in mm
            nty(double): negative position limit in Y in mm
            ntz(double): negative position limit in Z in mm
            nrx(double): negative position limit around the axis X in deg
            nry(double): negative position limit around the axis Y in deg
            nrz(double): negative position limit around the axis Y in deg

            ptx(double): positive position limit in X in mm
            pty(double): positive position limit in Y in mm
            ptz(double): positive position limit in Z in mm
            prx(double): positive position limit around the axis X in deg
            pry(double): positive position limit around the axis Y in deg
            prz(double): positive position limit around the axis Y in deg

        """

        raise NotImplementedError

    @dynamic_interface
    def set_default(self):
        """
        Restores the default configuration parameters. The command can be used to restore factory default parameters.
        The restored configuration is not automatically saved. refer to the command CFG_SAVE to save the parameters in
        order to keep them after a controller power off. The calculation of the hexapod position is suspended during the
        command execution.
        """

        raise NotImplementedError



import logging
from enum import Enum

# Default ports that are assigned to REQ-REP and PUB-SUB protocols of the registry services
DEFAULT_RS_REQ_PORT = 4242  # Handle requests
DEFAULT_RS_PUB_PORT = 4243  # Publish events
DEFAULT_RS_HB_PORT = 4244  # Heartbeats

DEFAULT_RS_DB_PATH = "service_registry.db"


class MessageType(Enum):
    """Message types using the envelope frame in the ROUTER-DEALER protocol."""

    REQUEST_WITH_REPLY = b"REQ"  # Client expects a reply
    REQUEST_NO_REPLY = b"NO-REQ"  # No reply expected by the client
    RESPONSE = b"REP"  # Response to a request
    NOTIFICATION = b"NOTIF"  # Server-initiated notification
    HEARTBEAT = b"HB"  # Heartbeat/health check


logger = logging.getLogger("egse.registry")


def is_service_registry_active(timeout: float = 0.5):
    """Check if the service registry is running and active.

    This function will send a 'health_check' request to the service registry and
    waits for the answer.

    If no reply was received after the given timeout [default=0.5s] the request
    will time out and return False.
    """

    from egse.registry.client import RegistryClient  # prevent circular import

    with RegistryClient(request_timeout=int(timeout * 1000)) as client:
        if not client.health_check():
            return False
        else:
            return True



"""
Notification Hub services for event-based messaging.

Provides asynchronous and synchronous publisher/subscriber classes for
sending and receiving events between core services via ZeroMQ.
"""

import asyncio
import json
import logging
from asyncio import CancelledError
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional

import zmq
import zmq.asyncio

from . import DEFAULT_COLLECTOR_PORT
from . import DEFAULT_PUBLISHER_PORT
from .event import NotificationEvent

NOTIFY_HUB_COLLECTOR = f"tcp://localhost:{DEFAULT_COLLECTOR_PORT}"
NOTIFY_HUB_PUBLISHER = f"tcp://localhost:{DEFAULT_PUBLISHER_PORT}"


class AsyncEventPublisher:
    """Publishes events asynchronously to the notification hub.

    Args:
        hub_address (str): Endpoint of the notification hub. [default=NOTIFY_HUB_COLLECTOR]
    """

    def __init__(self, hub_address: str = NOTIFY_HUB_COLLECTOR):
        self.hub_address = hub_address
        self.context = zmq.asyncio.Context()
        self.publisher = None
        self._connected = False
        self.logger = logging.getLogger("egse.event-pub")

    async def connect(self):
        """Connect to the notification hub"""
        if not self._connected:
            self.publisher = self.context.socket(zmq.PUSH)
            self.publisher.connect(self.hub_address)
            await asyncio.sleep(0.1)  # Allow connection to establish
            self._connected = True
            self.logger.info(f"Connected to hub at {self.hub_address}")

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    async def publish(self, event: NotificationEvent):
        """Publish an event"""

        await self.publisher.send(json.dumps(event.as_dict()).encode())
        self.logger.debug(f"Published: {event.event_type}")

    def disconnect(self):
        """Clean shutdown"""
        self._connected = False
        if self.publisher:
            # Keep the linger parameter, otherwise the event will not be sent when
            # you use a context manager to just send one event.
            self.publisher.close(linger=100)
        self.context.term()


class AsyncEventSubscriber:
    """Subscribes to events from the notification hub asynchronously.

    Args:
        subscriptions (List[str]): List of event types to subscribe to.
        hub_address (str): Endpoint of the notification hub. [default=NOTIFY_HUB_PUBLISHER]
    """

    def __init__(self, subscriptions: List[str], hub_address: str = NOTIFY_HUB_PUBLISHER):
        self.subscriptions = subscriptions
        self.hub_address = hub_address
        self.context = zmq.asyncio.Context()
        self.subscriber = None
        self.handlers: Dict[str, Callable] = {}
        self.running = False
        self.logger = logging.getLogger("egse.event-sub")

    async def connect(self):
        """Connect to the notification hub."""
        self.subscriber = self.context.socket(zmq.SUB)
        self.subscriber.connect(self.hub_address)

        # Subscribe to specific event types
        for event_type in self.subscriptions:
            self.subscriber.setsockopt(zmq.SUBSCRIBE, event_type.encode())
            self.logger.info(f"Subscribed to: {event_type}")

    def register_handler(self, event_type: str, handler: Callable):
        """Register a handler for an event type."""
        self.handlers[event_type] = handler

    async def start_listening(self):
        """Start listening for events"""

        self.running = True
        self.logger.info("Started listening for events")

        while self.running:
            try:
                if await self.subscriber.poll(timeout=1000):
                    topic, message_bytes = await self.subscriber.recv_multipart()

                    event_type = topic.decode()
                    event_data = json.loads(message_bytes.decode())

                    await self._handle_event(event_type, event_data)

                await asyncio.sleep(0.001)

            except Exception as exc:
                self.logger.error(f"Error receiving event: {exc}")
                await asyncio.sleep(1)
            except CancelledError:
                break

    async def _handle_event(self, event_type: str, event_data: Dict):
        """Handle received event"""
        handler = self.handlers.get(event_type)

        if handler:
            try:
                if asyncio.iscoroutinefunction(handler):
                    await handler(event_data)
                else:
                    await asyncio.get_event_loop().run_in_executor(None, handler, event_data)
            except Exception as exc:
                self.logger.error(f"Error handling {event_type}: {exc}")
        else:
            self.logger.warning(f"No handler registered for {event_type}")

    def disconnect(self):
        """Stop listening"""
        self.running = False
        if self.subscriber is not None:
            self.subscriber.close(linger=0)
        self.context.term()


class ServiceMessaging:
    """Convenience wrapper for publishing and subscribing to events.

    Args:
        service_name (str): Name of the service.
        subscriptions (List[str], optional): Event types to subscribe to.
    """

    def __init__(self, service_name: str, subscriptions: List[str] | None = None):
        self.service_name = service_name
        self.publisher = AsyncEventPublisher()
        self.subscriber = AsyncEventSubscriber(subscriptions or []) if subscriptions else None

    async def publish_event(self, event_type: str, data: Dict[Any, Any], correlation_id: Optional[str] = None):
        """Publish an event"""
        event = NotificationEvent(
            event_type=event_type,
            source_service=self.service_name,
            data=data,
            correlation_id=correlation_id,
        )
        await self.publisher.publish(event)

    def register_handler(self, event_type: str, handler: Callable):
        """Register event handler"""
        if self.subscriber:
            self.subscriber.register_handler(event_type, handler)

    async def start_listening(self):
        """Start listening for events"""
        if self.subscriber:
            await self.subscriber.start_listening()

    def disconnect(self):
        """Clean shutdown"""
        self.publisher.disconnect()
        if self.subscriber:
            self.subscriber.disconnect()


class EventPublisher:
    """Publishes events synchronously to the notification hub.

    Args:
        hub_address (str): Endpoint of the notification hub. [default=NOTIFY_HUB_COLLECTOR]
    """

    def __init__(self, hub_address: str = NOTIFY_HUB_COLLECTOR):
        self.hub_address = hub_address
        self.context = zmq.Context().instance()
        self.publisher = None
        self._connected = False
        self.logger = logging.getLogger("egse.event-pub")

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def connect(self):
        """Connect to the notification hub"""
        if not self._connected:
            self.publisher = self.context.socket(zmq.PUSH)
            self.publisher.connect(self.hub_address)
            # time.sleep(0.1)  # Allow connection to establish
            self._connected = True
            self.logger.info(f"Connected to hub at {self.hub_address}")

    def publish(self, event: NotificationEvent):
        """Publish an event"""
        self.publisher.send(json.dumps(event.as_dict()).encode())
        self.logger.debug(f"Published: {event.event_type} with correlation_id={event.correlation_id}")

    def disconnect(self):
        """Clean shutdown"""
        if self._connected:
            # Keep the linger parameter, otherwise the event will not be sent when
            # you use the context manager to just send one event.
            self.publisher.close(linger=100)
            self._connected = False


class EventSubscriber:
    """Subscribes to events from the notification hub. Use this class in a synchronous context.

    Args:
        subscriptions (List[str]): List of event types to subscribe to.
        hub_address (str): Endpoint of the notification hub. [default = NOTIFY_HUB_COLLECTOR]
    """

    def __init__(self, subscriptions: List[str], hub_address: str = NOTIFY_HUB_PUBLISHER):
        self.subscriptions = subscriptions
        self.hub_address = hub_address
        self.context = zmq.Context().instance()
        self.subscriber = None
        self.handlers: Dict[str, Callable] = {}
        self.logger = logging.getLogger("egse.event-sub")

    def connect(self):
        """Connect to the notification hub"""
        self.subscriber = self.context.socket(zmq.SUB)
        self.subscriber.connect(self.hub_address)

        # Subscribe to specific event types
        for event_type in self.subscriptions:
            self.subscriber.setsockopt(zmq.SUBSCRIBE, event_type.encode())
            self.logger.info(f"Subscribed to: {event_type or 'all'}")

    def register_handler(self, event_type: str, handler: Callable):
        """Register a handler for an event type"""
        self.handlers[event_type] = handler

    @property
    def socket(self):
        return self.subscriber

    def poll(self, timeout=1000):
        return self.subscriber.poll(timeout=timeout)

    def handle_event(self, return_event_data: bool = False) -> dict:
        topic, message_bytes = self.subscriber.recv_multipart()

        event_type = topic.decode()
        event_data = json.loads(message_bytes.decode())

        if return_event_data:
            return event_data
        else:
            self._handle_event(event_type, event_data)
            return {}

    def _handle_event(self, event_type: str, event_data: Dict):
        """Handle received event"""
        handler = self.handlers.get(event_type)

        if handler:
            try:
                handler(event_data)
            except Exception as exc:
                self.logger.error(f"Error handling {event_type} by {handler.__name__}: {exc}", exc_info=True)
        else:
            self.logger.warning(f"No handler registered for {event_type}")

    def disconnect(self):
        """Stop listening"""
        if self.subscriber:
            self.subscriber.close(linger=0)



import pickle

from egse.zmq_ser import zmq_error_response
from egse.zmq_ser import zmq_json_request
from egse.zmq_ser import zmq_json_response
from egse.zmq_ser import zmq_string_request
from egse.zmq_ser import zmq_string_response


def test_zmq_requests():
    assert zmq_string_request("") == [b"MESSAGE_TYPE:STRING", b""]
    assert zmq_string_request("hello") == [b"MESSAGE_TYPE:STRING", b"hello"]

    assert zmq_json_request({}) == [b"MESSAGE_TYPE:JSON", b"{}"]
    assert zmq_json_request({"command": "do_something"}) == [b"MESSAGE_TYPE:JSON", b'{"command": "do_something"}']


def test_zmq_responses():
    assert zmq_string_response("") == [b"MESSAGE_TYPE:STRING", b""]
    assert zmq_string_response("Good Job!") == [b"MESSAGE_TYPE:STRING", b"Good Job!"]

    assert zmq_json_response({}) == [b"MESSAGE_TYPE:JSON", b"{}"]
    assert zmq_json_response({"success": True, "message": "all good", "metadata": {"version": "0.1.0"}}) == [
        b"MESSAGE_TYPE:JSON",
        b'{"success": true, "message": "all good", "metadata": {"version": "0.1.0"}}',
    ]
    assert zmq_json_response(
        {
            "success": True,
            "message": {"name": "cgse", "description": "Great software"},
            "metadata": {"version": "0.1.0"},
        }
    ) == [
        b"MESSAGE_TYPE:JSON",
        b"{"
        b'"success": true, '
        b'"message": {"name": "cgse", "description": "Great software"}, "metadata": {"version": "0.1.0"}'
        b"}",
    ]


def test_zmq_error_responses():
    assert zmq_error_response({}) == [b"MESSAGE_TYPE:ERROR", b"\x80\x04}\x94."]
    response = zmq_error_response(
        {
            "success": False,
            "message": "Incorrect message type: MESSAGE_TYPE:PICKLE",
            "metadata": {
                "data": {
                    "file": "/Users/rik/github/cgse/libs/cgse-common/src/egse/async_control.py",
                    "lineno": 322,
                    "function": "_process_device_command",
                },
            },
        }
    )
    assert isinstance(response, list)
    assert len(response) == 2
    assert response[0] == b"MESSAGE_TYPE:ERROR"
    assert isinstance(response[1], bytes)
    assert pickle.loads(response[1]) == {
        "success": False,
        "message": "Incorrect message type: MESSAGE_TYPE:PICKLE",
        "metadata": {
            "data": {
                "file": "/Users/rik/github/cgse/libs/cgse-common/src/egse/async_control.py",
                "lineno": 322,
                "function": "_process_device_command",
            },
        },
    }



# Provide basic information about the egse logger that is of interest to the developer.
# Do not assume the log_cs is running, this function shall also provide the information
# even if the logger is not running.
#
# usage:
#   $ python -m egse.logger
import logging
import sys

import rich

from egse.env import get_log_file_location


def main():
    rich.print(f"Log file location: {get_log_file_location()}")


def test():
    # Any logger name that starts with "egse." will also be logged by the log_cs.

    logger = logging.getLogger("egse.0mq-log-test")
    logger.debug("Hello, ZeroMQ logging: This is a DEBUG message.")
    logger.info("Hello, ZeroMQ logging: This is an INFO message.")
    logger.warning("Hello, ZeroMQ logging: This is a WARNING message.")
    logger.error("Hello, ZeroMQ logging: This is an ERROR message.")
    logger.critical("Hello, ZeroMQ logging: This is a CRITICAL message.")
    try:
        raise ValueError("A fake ValueError, raised for testing.")
    except ValueError:
        logger.exception("Hello, ZeroMQ logging: This is an EXCEPTION message.")

    # The following log messages will not be logged by log_cs.

    logger = logging.getLogger("plain-log-test")

    logger.debug("Vanilla logging: This is a DEBUG message.")
    logger.info("Vanilla logging: This is an INFO message.")
    logger.warning("Vanilla logging: This is a WARNING message.")
    logger.error("Vanilla logging: This is an ERROR message.")
    logger.critical("Vanilla logging: This is a CRITICAL message.")
    try:
        raise ValueError("A fake ValueError, raised for testing.")
    except ValueError:
        logger.exception("Vanilla logging: This is an EXCEPTION message.")


if __name__ == "__main__":
    # test()

    sys.exit(main())



import logging
from pathlib import Path

from egse.control import ControlServer
from egse.protocol import CommandProtocol
from egse.settings import Settings
from egse.setup import load_setup
from egse.system import format_datetime
from egse.tempcontrol.lakeshore.lakeshore336 import LakeShore336Interface, LakeShore336Simulator, LakeShore336Controller
from egse.tempcontrol.lakeshore.lakeshore336_devif import LakeShore336Command
from egse.zmq_ser import bind_address

logger = logging.getLogger(__name__)
_HERE = Path(__file__).parent
COMMAND_SETTINGS = Settings.load(filename="lakeshore336.yaml", location=_HERE)
CTRL_SETTINGS = Settings.load("LakeShore336 Control Server")
SITE_ID = Settings.load("SITE").ID


class LakeShore336Protocol(CommandProtocol):
    def __init__(self, control_server: ControlServer, device_id: str, simulator: bool = False):
        super().__init__(control_server)
        self.device_id = device_id
        self.simulator = simulator

        setup = load_setup()
        self.input_channels = setup.gse.tempcontrol.lakeshore336[device_id]["input_channels"]
        if isinstance(self.input_channels, str):
            self.input_channels = [self.input_channels]
        self.output_channels = setup.gse.tempcontrol.lakeshore336[device_id]["output_channels"]
        if isinstance(self.output_channels, int):
            self.output_channels = [self.output_channels]

        if self.simulator:
            self.lakeshore: LakeShore336Interface = LakeShore336Simulator(device_id)
        else:
            self.lakeshore: LakeShore336Interface = LakeShore336Controller(device_id)

        self.storage_mnemonic = self.control_server.get_storage_mnemonic()

        self.load_commands(COMMAND_SETTINGS.Commands, LakeShore336Command, LakeShore336Interface)

        self.build_device_method_lookup_table(self.lakeshore)

        # self.synoptics = SynopticsManagerProxy()

    def get_bind_address(self):
        return bind_address(
            self.control_server.get_communication_protocol(),
            self.control_server.get_commanding_port(),
        )

    def get_status(self):
        return super().get_status()

    def get_housekeeping(self) -> dict:
        hk_dict = dict()

        try:
            # Temperature sensors (input)

            for input_channel in self.input_channels:
                hk_dict[f"G{SITE_ID}_{self.storage_mnemonic}_T_{input_channel}"] = self.lakeshore.get_temperature(
                    input_channel=input_channel
                )

            # Heaters (output)

            # for output_channel in self.output_channels:
            for output_channel in self.output_channels:  # list(set(self.output_channels) & {1, 2}):
                if output_channel in {1, 2}:
                    pid_parameters = self.lakeshore.get_pid_parameters(output_channel)
                    hk_dict[f"G{SITE_ID}_{self.storage_mnemonic}_P_{output_channel}"] = pid_parameters[0]
                    hk_dict[f"G{SITE_ID}_{self.storage_mnemonic}_I_{output_channel}"] = pid_parameters[1]
                    hk_dict[f"G{SITE_ID}_{self.storage_mnemonic}_D_{output_channel}"] = pid_parameters[2]

                    hk_dict[f"G{SITE_ID}_{self.storage_mnemonic}_HEATER_{output_channel}"] = (
                        self.lakeshore.get_heater_output(output_channel)
                    )

                hk_dict[f"G{SITE_ID}_{self.storage_mnemonic}_SETPOINT_{output_channel}"] = (
                    self.lakeshore.get_temperature_setpoint(output_channel)
                )

            # hk_dict[f"G{SITE_ID}_{self.storage_mnemonic}_TEMP_A"] = self.lakeshore.get_temperature()
            # pid_params = self.lakeshore.get_pid_parameters(1)
            # hk_dict[f"G{SITE_ID}_{self.storage_mnemonic}_P_VALUE"] = pid_params[0]
            # hk_dict[f"G{SITE_ID}_{self.storage_mnemonic}_I_VALUE"] = pid_params[1]
            # hk_dict[f"G{SITE_ID}_{self.storage_mnemonic}_D_VALUE"] = pid_params[2]
            # hk_dict[f"G{SITE_ID}_{self.storage_mnemonic}_HEATER_VALUE"] = self.lakeshore.get_heater_output(
            #     output_channel=1)
            # hk_dict[f"G{SITE_ID}_{self.storage_mnemonic}_SETPOINT_VALUE"] = self.lakeshore.get_temperature_setpoint(
            #     output_channel=1)

            # for hk_name in metrics_dict.keys():
            #     index_lsci = hk_name.split("_")
            #     if (len(index_lsci) > 2):
            #         if int(index_lsci[2]) == int(self.device_id):
            #             metrics_dict[hk_name].set(hk_dict[hk_name])

            # # Send the HK acquired so far to the Synoptics Manager
            # self.synoptics.store_th_synoptics(hk_dict)
        except Exception as exc:
            logger.warning(f"failed to get HK ({exc})")
        hk_dict["timestamp"] = format_datetime()

        return hk_dict

    def quit(self):
        self.lakeshore.quit()



import subprocess
import sys
import textwrap
from pathlib import Path
from typing import Annotated
from typing import TextIO

import rich
import typer

from egse.env import get_log_file_location
from egse.system import all_logging_disabled

puna = typer.Typer(name="puna", help="PUNA Positioning Hexapod, Symétrie")

zonda = typer.Typer(name="zonda", help="ZONDA Positioning Hexapod, Symétrie")

joran = typer.Typer(name="joran", help="JORAN Positioning Hexapod, Symétrie")


def redirect_output_to(output_fn: str) -> TextIO:
    """Open file in the log folder where process output will be redirected."""

    location = get_log_file_location()
    output_path = Path(location, output_fn).expanduser()

    rich.print(f"Output will be redirected to {output_path!s}")

    out = open(output_path, "w")

    return out


def start_hexapod_cs_process(device_name, device_id, simulator):
    """Generic function to start the hexapod control server in the background."""

    rich.print(f"Starting the {device_name} hexapod control server for {device_id} – {simulator = }")

    out = redirect_output_to(f"{device_name.lower()}_cs.{device_id.lower()}.start.out")

    cmd = [sys.executable, "-m", f"egse.hexapod.symetrie.{device_name.lower()}_cs", "start", device_id]
    if simulator:
        cmd.append("--simulator")

    subprocess.Popen(cmd, stdout=out, stderr=out, stdin=subprocess.DEVNULL, close_fds=True)


def stop_hexapod_cs_process(device_name, device_id):
    """Generic function to stop the hexapod control server in the background."""

    rich.print(f"Terminating hexapod {device_name} control server for {device_id}...")

    out = redirect_output_to(f"{device_name.lower()}_cs.{device_id.lower()}.stop.out")

    cmd = [sys.executable, "-m", f"egse.hexapod.symetrie.{device_name.lower()}_cs", "stop", device_id]

    subprocess.Popen(cmd, stdout=out, stderr=out, stdin=subprocess.DEVNULL, close_fds=True)


# ---------- PUNA Commands ---------------------------------------------------------------------------------------------


@puna.command(name="start")
def start_puna(
    device_id: Annotated[str, typer.Argument(help="the device identifier, identifies the hardware controller")],
    simulator: Annotated[
        bool, typer.Option("--simulator", "--sim", help="use a device simulator as the backend")
    ] = False,
):
    """
    Start the PUNA hexapod control server. The control server is always started in the background.
    """
    start_hexapod_cs_process("PUNA", device_id, simulator)


@puna.command(name="stop")
def stop_puna(
    device_id: Annotated[str, typer.Argument(help="the device identifier, identifies the hardware controller")],
):
    """Stop the PUNA hexapod control server."""

    stop_hexapod_cs_process("PUNA", device_id)


@puna.command(name="status")
def status_puna(device_id: str):
    """Print status information on the PUNA hexapod control server."""

    with all_logging_disabled():
        from egse.hexapod.symetrie import puna_cs

        puna_cs.status(device_id)


@puna.command(name="start-sim")
def start_puna_sim(device_id: str):
    """Start the PUNA Hexapod Simulator (the device simulator)."""

    rich.print(
        textwrap.dedent(
            f"""\
            [orange3]The PUNA simulator is in development, for now use the `--sim` option when starting the control 
            server.[/]
            
            The `--sim` option will use a Controller clas that doesn't send commands to the device, but simulates
            the requests by performing actions on the reference frames defined in the hexapod. This means you are 
            not exercising the actual device commanding, but the net result or outcome is the same.
            
            usage: cgse puna start --sim {device_id}
            """
        )
    )
    return

    rich.print("Starting service PUNA Simulator")

    out = redirect_output_to(f"puna_sim.{device_id.lower()}.start.out")

    subprocess.Popen(
        [sys.executable, "-m", "egse.hexapod.symetrie.puna_sim", "start", device_id],
        stdout=out,
        stderr=out,
        stdin=subprocess.DEVNULL,
        close_fds=True,
    )


@puna.command(name="stop-sim")
def stop_puna_sim(device_id: str):
    """Stop the PUNA Hexapod Simulator."""
    rich.print("Terminating the PUNA simulator.")

    out = redirect_output_to(f"puna_sim.{device_id.lower()}.stop.out")

    subprocess.Popen(
        [sys.executable, "-m", "egse.hexapod.symetrie.puna_sim", "stop", device_id],
        stdout=out,
        stderr=out,
        stdin=subprocess.DEVNULL,
        close_fds=True,
    )


# ---------- ZONDA Commands --------------------------------------------------------------------------------------------


@zonda.command(name="start")
def start_zonda(
    device_id: Annotated[str, typer.Argument(help="the device identifier, identifies the hardware controller")],
    simulator: Annotated[
        bool, typer.Option("--simulator", "--sim", help="use a device simulator as the backend")
    ] = False,
):
    """
    Start the ZONDA hexapod control server. The control server is always started in the background.
    """

    start_hexapod_cs_process("ZONDA", device_id, simulator)


@zonda.command(name="stop")
def stop_zonda(
    device_id: Annotated[str, typer.Argument(help="the device identifier, identifies the hardware controller")],
):
    """Stop the ZONDA hexapod control server."""

    stop_hexapod_cs_process("ZONDA", device_id)


@zonda.command(name="status")
def status_zonda(device_id: str):
    """Print status information on the ZONDA hexapod control server."""

    with all_logging_disabled():
        from egse.hexapod.symetrie import zonda_cs

        zonda_cs.status(device_id)
